/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::LVL1Controller
 * @author     Raphael ( ufdrv at student dot kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "LVL1Controller.h"

#include <ArmarXCore/core/util/algorithm.h>

#include "../RobotUnit.h"

using namespace armarx;

const LVL1ControllerPtr LVL1Controller::NullPtr
{
    nullptr
};

bool LVL1Controller::isControllerActive(const Ice::Current&) const
{
    return isActive;
}
bool LVL1Controller::isControllerRequested(const Ice::Current&) const
{
    return isRequested;
}
bool LVL1Controller::isDeletable(const Ice::Current&) const
{
    return deletable;
}

bool LVL1Controller::hasControllerError(const Ice::Current&) const
{
    return deactivatedBecauseOfError;
}

std::string LVL1Controller::getInstanceName(const Ice::Current&) const
{
    return getName();
}

LVL1ControllerDescription LVL1Controller::getControllerDescription(const Ice::Current&) const
{
    LVL1ControllerDescription d;
    d.className = getClassName();
    ARMARX_CHECK_EXPRESSION(getProxy(-1));
    d.controller = LVL1ControllerInterfacePrx::uncheckedCast(getProxy(-1));
    d.controlModeAssignment = controlDeviceControlModeMap;
    d.instanceName = getInstanceName();
    d.deletable = deletable;
    return d;
}

StringStringDictionary LVL1Controller::getControlDeviceUsedControlModeMap(const Ice::Current&) const
{
    return controlDeviceControlModeMap;
}

boost::optional<std::vector<char> > LVL1Controller::isNotInConflictWith(const std::vector<char>& used) const
{
    ARMARX_CHECK_EXPRESSION(used.size() == controlDeviceUsedBitmap.size());
    auto result = used;
    for (std::size_t i = 0; i < controlDeviceUsedBitmap.size(); ++i)
    {
        if (controlDeviceUsedBitmap.at(i))
        {
            if (used.at(i))
            {
                return boost::optional<std::vector<char>>();
            }
            result.at(i) = true;
        }

    }
    return {true, std::move(result)};
}

LVL1ControllerStatus LVL1Controller::getControllerStatus(const Ice::Current&) const
{
    LVL1ControllerStatus s;
    s.active = isActive;
    s.instanceName = getInstanceName();
    s.requested = isRequested;
    s.error = deactivatedBecauseOfError;
    s.timestampUSec = std::chrono::duration_cast<std::chrono::microseconds>(std::chrono::high_resolution_clock::now().time_since_epoch()).count();
    return s;
}

LVL1ControllerDescriptionWithStatus LVL1Controller::getControllerDescriptionWithStatus(const Ice::Current&) const
{
    LVL1ControllerDescriptionWithStatus ds;
    ds.description = getControllerDescription();
    ds.status = getControllerStatus();
    return ds;
}

RobotUnitInterfacePrx LVL1Controller::getRobotUnit(const Ice::Current&) const
{
    ARMARX_CHECK_EXPRESSION(robotUnit);
    return RobotUnitInterfacePrx::uncheckedCast(robotUnit->getProxy(-1));
}

void LVL1Controller::activateController(const Ice::Current&) const
{
    ARMARX_CHECK_EXPRESSION(robotUnit);
    robotUnit->activateLVL1Controller(getInstanceName());
}

void LVL1Controller::deactivateController(const Ice::Current&) const
{
    ARMARX_CHECK_EXPRESSION(robotUnit);
    robotUnit->deactivateLVL1Controller(getInstanceName());
}

void LVL1Controller::deleteController(const Ice::Current&) const
{
    ARMARX_CHECK_EXPRESSION(robotUnit);
    robotUnit->deleteLVL1Controller(getInstanceName());
}

void LVL1Controller::publish(const SensorAndControl& sac, const DebugDrawerInterfacePrx& draw, const DebugObserverInterfacePrx& observer)
{
    const bool active = isActive;
    if (publishActive.exchange(active) != active)
    {
        if (active)
        {
            ARMARX_DEBUG << "activating publishing for " << getInstanceName();
            onPublishActivation(draw, observer);
        }
        else
        {
            ARMARX_DEBUG << "deactivating publishing for " << getInstanceName();
            onPublishDeactivation(draw, observer);
        }
    }
    if (active)
    {
        onPublish(sac, draw, observer);
    }
}

void LVL1Controller::deactivatePublish(const DebugDrawerInterfacePrx& draw, const DebugObserverInterfacePrx& observer)
{
    if (publishActive.exchange(false))
    {
        ARMARX_DEBUG << "forced deactivation of publishing for " << getInstanceName();
        onPublishDeactivation(draw, observer);
    }
}

void LVL1Controller::rtActivateController()
{
    if (!isActive)
    {
        deactivatedBecauseOfError = false;
        ARMARX_INFO << "activating lvl1controller " << getInstanceName() << " " << this;
        rtPreActivateController();
        isActive = true;
    }
    else
    {
        ARMARX_INFO << "lvl1controller " << getInstanceName() << " is already active " << this;
    }
}

void LVL1Controller::rtDeactivateController()
{
    if (isActive)
    {
        isActive = false;
        ARMARX_INFO << "deactivating lvl1controller " << getInstanceName() << " " << this;
        rtPostDeactivateController();
    }
    else
    {
        ARMARX_INFO << "lvl1controller " << getInstanceName() << " is already deactivated " << this;
    }
}

void LVL1Controller::rtDeactivateControllerBecauseOfError()
{
    deactivatedBecauseOfError = true;
    rtDeactivateController();
}

void LVL1Controller::robotUnitInit(
    RobotUnit* ru,
    std::size_t ctrlId,
    StringStringDictionary ctrlDeviceControlModeMap,
    std::vector<char> ctrlDeviceUsedBitmap,
    std::vector<std::size_t> ctrlDeviceUsedIndices,
    bool ctrlDeletable)
{
    ARMARX_CHECK_EXPRESSION(ru);
    ARMARX_CHECK_EXPRESSION(ctrlDeviceControlModeMap.size());
    ARMARX_CHECK_EXPRESSION(ctrlId != std::numeric_limits<std::size_t>::max());
    ARMARX_CHECK_EXPRESSION_W_HINT(ru->getNumberOfControlDevices() == ctrlDeviceUsedBitmap.size(),
                                   "ru->getNumberOfControlDevices() == ctrlDeviceUsedBitmap.size() ->" +
                                   std::to_string(ru->getNumberOfControlDevices()) + " > " + std::to_string(ctrlDeviceUsedBitmap.size()));
    ARMARX_CHECK_EXPRESSION(ctrlDeviceControlModeMap.size() == ctrlDeviceUsedIndices.size());

    robotUnit = ru;
    id = ctrlId;
    controlDeviceControlModeMap = std::move(ctrlDeviceControlModeMap);
    controlDeviceUsedBitmap     = std::move(ctrlDeviceUsedBitmap);
    controlDeviceUsedIndices    = std::move(ctrlDeviceUsedIndices);
    deletable = ctrlDeletable;
}

