/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::LVL1Controller
 * @author     Raphael ( ufdrv at student dot kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#ifndef _ARMARX_LIB_RobotAPI_LVL1Controller_HPP
#define _ARMARX_LIB_RobotAPI_LVL1Controller_HPP

#include <ArmarXCore/core/util/TemplateMetaProgramming.h>

namespace armarx
{
    template<class ItT>
    boost::optional<std::vector<char>> LVL1Controller::AreNotInConflict(ItT first, ItT last)
    {
        if (first == last)
        {
            return {true, std::vector<char>{}};
        }
        std::size_t n = (*first)->getControlDeviceUsedBitmap().size();
        std::vector<char> inuse(n, false);
        while (first != last)
        {
            auto r = (*first)->isNotInConflictWith(inuse);
            if (!r)
            {
                return r;
            }
            inuse = std::move(r.get());
            ++first;
        }
        return {true, std::move(inuse)};
    }

    namespace detail
    {
        template<class LVL1ControllerT, class = void, class = void>
        class LVL1ControllerRegistryEntryHelper : public LVL1ControllerRegistryEntry
        {
            virtual LVL1ControllerPtr create(LVL1ControllerDescriptionProviderInterfacePtr prov, const LVL1ControllerConfigPtr& config, const boost::shared_ptr<VirtualRobot::Robot>& rob) const final
            {
                ARMARX_CHECK_EXPRESSION_W_HINT(prov, "DataProvider is NULL!");
                return new LVL1ControllerT(prov, config, rob);
            }
            virtual WidgetDescription::WidgetPtr GenerateConfigDescription(const boost::shared_ptr<VirtualRobot::Robot>&, const std::map<std::string, ConstControlDevicePtr>&, const std::map<std::string, ConstSensorDevicePtr>&) const final
            {
                ARMARX_CHECK_EXPRESSION(!"This function should never be called");
            }
            virtual LVL1ControllerConfigPtr GenerateConfigFromVariants(const StringVariantBaseMap&) const
            {
                ARMARX_CHECK_EXPRESSION(!"This function should never be called");
            }
            virtual bool hasRemoteConfiguration() const
            {
                return false;
            }
        };

        ARMARX_META_MAKE_HAS_MEMBER_FNC_CHECK(hasGenerateConfigDescription,  GenerateConfigDescription,  LVL1Controller::GenerateConfigDescriptionFunctionSignature);
        ARMARX_META_MAKE_HAS_MEMBER_FNC_CHECK(hasGenerateConfigFromVariants, GenerateConfigFromVariants, LVL1Controller::GenerateConfigFromVariantsFunctionSignature);

#define ARMARX_ASSERT_LVL1CONTROLLER_HAS_CONSTRUCTION_GUI(T)            \
    static_assert                                                       \
    (                                                                   \
            ::armarx::detail::hasGenerateConfigDescription<T>::value && \
            ::armarx::detail::hasGenerateConfigFromVariants<T>::value,  \
            #T " does not offer a construction gui"                     \
    )

        template<class LVL1ControllerT>
        class LVL1ControllerRegistryEntryHelper <
            LVL1ControllerT,
            //            meta::void_t<decltype(LVL1ControllerT::GenerateConfigDescription)>,
            typename std::enable_if < hasGenerateConfigDescription<LVL1ControllerT>::value >::type,
            //            meta::void_t<decltype(LVL1ControllerT::GenerateConfigFromVariants)
            typename std::enable_if < hasGenerateConfigFromVariants<LVL1ControllerT>::value >::type
            > : public LVL1ControllerRegistryEntry
        {
            virtual LVL1ControllerPtr create(LVL1ControllerDescriptionProviderInterfacePtr prov, const LVL1ControllerConfigPtr& config, const boost::shared_ptr<VirtualRobot::Robot>& rob) const final
            {
                ARMARX_CHECK_EXPRESSION_W_HINT(prov, "DataProvider is NULL!");
                return new LVL1ControllerT(prov, config, rob);
            }
            virtual WidgetDescription::WidgetPtr GenerateConfigDescription(
                const boost::shared_ptr<VirtualRobot::Robot>& robot,
                const std::map<std::string, ConstControlDevicePtr>& controlDevices,
                const std::map<std::string, ConstSensorDevicePtr>& sensorDevices
            ) const final
            {
                WidgetDescription::WidgetPtr cfg = LVL1ControllerT::GenerateConfigDescription(robot, controlDevices, sensorDevices);
                if (!cfg)
                {
                    cfg = new WidgetDescription::Widget;
                }
                return cfg;
            }
            virtual LVL1ControllerConfigPtr GenerateConfigFromVariants(const StringVariantBaseMap& variants) const
            {
                return LVL1ControllerT::GenerateConfigFromVariants(variants);
            }
            virtual bool hasRemoteConfiguration() const
            {
                return true;
            }
        };
    }

    using LVL1ControllerRegistry = Registrar<std::unique_ptr<LVL1ControllerRegistryEntry>>;
    template<class ControllerType>
    struct LVL1ControllerRegistration
    {
        LVL1ControllerRegistration(const std::string& name)
        {
            LVL1ControllerRegistry::registerElement(name, std::unique_ptr<LVL1ControllerRegistryEntry>(new detail::LVL1ControllerRegistryEntryHelper<ControllerType>));
        }
    };
}
#endif
