/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::RobotUnit
 * @author     Raphael ( ufdrv at student dot kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "RobotUnit.h"

#include <algorithm>

#include <VirtualRobot/XML/RobotIO.h>
#include <VirtualRobot/Nodes/ForceTorqueSensor.h>
#include <VirtualRobot/Robot.h>
#include <VirtualRobot/RobotNodeSet.h>

#include <ArmarXCore/core/system/cmake/CMakePackageFinder.h>
#include <ArmarXCore/core/system/DynamicLibrary.h>
#include <ArmarXCore/observers/variant/TimestampVariant.h>

#include "Units/ForceTorqueSubUnit.h"
#include "Units/InertialMeasurementSubUnit.h"
#include "Units/KinematicSubUnit.h"
#include "Units/PlatformSubUnit.h"


namespace armarx
{
    namespace detail
    {
        template<class TimeT = std::chrono::microseconds>
        struct TimerTag
        {
            //assume it is std::chrono

            using ClockT = std::chrono::high_resolution_clock;
            const ClockT::time_point beg;

            TimerTag() : beg {ClockT::now()} {}

            TimeT stop()
            {
                return std::chrono::duration_cast<TimeT>(ClockT::now() - beg);
            }
        };

        template<>
        struct TimerTag<TimestampVariant> : TimerTag<>
        {
            TimestampVariant stop()
            {
                return {TimerTag<std::chrono::microseconds>::stop().count()};
            }
        };

        template<>
        struct TimerTag<IceUtil::Time> : TimerTag<>
        {
            TimestampVariant stop()
            {
                return IceUtil::Time::microSeconds(TimerTag<std::chrono::microseconds>::stop().count());
            }
        };

        template <class Fun, class TimeT>
        TimeT operator*(TimerTag<TimeT>&& t, Fun&& fn)
        {
            fn();
            return t.stop();
        }

        //for virtual time

        template<class TimeT = IceUtil::Time>
        struct VirtualTimerTag;

        template<>
        struct VirtualTimerTag<TimestampVariant>
        {
            const IceUtil::Time beg;
            VirtualTimerTag() : beg {TimeUtil::GetTime()} {}
            TimestampVariant stop()
            {
                return {TimeUtil::GetTime() - beg};
            }
        };

        template<>
        struct VirtualTimerTag<IceUtil::Time>
        {
            const IceUtil::Time beg;
            VirtualTimerTag() : beg {TimeUtil::GetTime()} {}
            TimestampVariant stop()
            {
                return TimeUtil::GetTime() - beg;
            }
        };

        template <class Fun, class TimeT>
        TimeT operator*(VirtualTimerTag<TimeT>&& t, Fun&& fn)
        {
            fn();
            return t.stop();
        }
    }
}
#define ARMARX_TIMER(...) ::armarx::detail::TimerTag<__VA_ARGS__>{} *[&]
#define ARMARX_VIRTUAL_TIMER(...) ::armarx::detail::VirtualTimerTag<__VA_ARGS__>{} *[&]


namespace std
{
    std::string to_string(armarx::RobotUnitState s)
    {
        switch (s)
        {
            case armarx::RobotUnitState::PreComponentInitialization:
                return "PreComponentInitialization";
            case armarx::RobotUnitState::InitializingDevices:
                return "InitializingDevices";
            case armarx::RobotUnitState::InitializingUnits:
                return "InitializingUnits";
            case armarx::RobotUnitState::WaitingForRTThreadInitialization:
                return "WaitingForRTThreadInitialization";
            case armarx::RobotUnitState::Running:
                return "Running";
            case armarx::RobotUnitState::Exiting:
                return "Exiting";
        }
        throw std::invalid_argument {"Unknown state " + std::to_string(static_cast<std::size_t>(s))};
    }
}

namespace armarx
{

    namespace detail
    {
        std::function<std::string(const std::string&)> makeDebugObserverNameFixer(const std::string& prefix)
        {
            return
                [prefix](const std::string & name)
            {
                std::string s = prefix + name;
                std::replace(s.begin(), s.end(), '.', '_');
                std::replace(s.begin(), s.end(), ':', '_');
                return s;
            };
        }
    }


    const std::string RobotUnit::rtThreadTimingsSensorDeviceName = "RTThreadTimings";

    const std::set<RobotUnitState> RobotUnit::devicesReadyStates {RobotUnitState::InitializingUnits, RobotUnitState::WaitingForRTThreadInitialization, RobotUnitState::Running};
    RobotUnit::GuardType RobotUnit::getGuard() const
    {
        ARMARX_CHECK_EXPRESSION_W_HINT(std::this_thread::get_id() != rtThreadId, "Atempted to lock mutex in RT thread");
        return GuardType {dataMutex};
    }

    std::chrono::microseconds RobotUnit::MicroNow()
    {
        return std::chrono::duration_cast<std::chrono::microseconds>(std::chrono::high_resolution_clock::now().time_since_epoch());
    }

    void RobotUnit::checkLVL1ControllerClassName(const std::string& className) const
    {
        if (!LVL1ControllerRegistry::has(className))
        {
            std::stringstream ss;
            ss << "Requested controller class '" << className << "' unknown! Known classes:" << LVL1ControllerRegistry::getKeys()
               << " (If this class exists in a different lib then load it via loadLibFromPath(path) or loadLibFromPackage(package, lib))";
            ARMARX_ERROR << ss.str();
            throw InvalidArgumentException {ss.str()};
        }
    }

    void RobotUnit::addControlDevice(const ControlDevicePtr& cd)
    {
        auto guard = getGuard();
        throwIfStateIsNot(RobotUnitState::InitializingDevices, __FUNCTION__);
        controlDevices.add(cd->getDeviceName(), cd);
        ARMARX_INFO << "added ControlDevice " << cd->getDeviceName();
    }

    void RobotUnit::addSensorDevice(const SensorDevicePtr& sd)
    {
        auto guard = getGuard();
        throwIfStateIsNot(RobotUnitState::InitializingDevices, __FUNCTION__);
        if (!sd)
        {
            std::stringstream ss;
            ss << "RobotUnit::addSensorDevice: SensorDevice is nullptr";
            ARMARX_ERROR << ss.str();
            throw InvalidArgumentException {ss.str()};
        }
        if (!sd->getSensorValue())
        {
            std::stringstream ss;
            ss << "RobotUnit::addSensorDevice: SensorDevice " << sd->getDeviceName() << " has null SensorValue (this is not allowed)";
            ARMARX_ERROR << ss.str();
            throw InvalidArgumentException {ss.str()};
        }
        if (sd->getDeviceName() == rtThreadTimingsSensorDeviceName)
        {
            if (!std::dynamic_pointer_cast<RTThreadTimingsSensorDevice>(sd))
            {
                throw InvalidArgumentException
                {
                    "You tried to add a SensorDevice with the name " + sd->getDeviceName() + " which does not derive from RTThreadTimingsSensorDevice. (Don't do this)"
                };
            }
            sensorDevices.add(sd->getDeviceName(), sd); //this checks if we already added such a device (do this before setting timingSensorDevice)
            rtThreadTimingsSensorDevice = std::dynamic_pointer_cast<RTThreadTimingsSensorDevice>(sd);
        }
        else
        {
            sensorDevices.add(sd->getDeviceName(), sd);
        }
        ARMARX_INFO << "added SensorDevice " << sd->getDeviceName() << " (valuetype = " << sd->getSensorValueType() << ")";
    }

    RTThreadTimingsSensorDevicePtr RobotUnit::createRTThreadTimingSensorDevice() const
    {
        return std::make_shared<RTThreadTimingsSensorDeviceImpl<>>(rtThreadTimingsSensorDeviceName);
    }

    void RobotUnit::finishDeviceInitialization()
    {
        auto beg = MicroNow();
        auto guard = getGuard();
        throwIfStateIsNot(RobotUnitState::InitializingDevices, __FUNCTION__);
        if (!sensorDevices.has(rtThreadTimingsSensorDeviceName))
        {
            addSensorDevice(createRTThreadTimingSensorDevice());
        }
        state = RobotUnitState::InitializingUnits;
        //check added devices
        ARMARX_DEBUG << "checking ControlDevices:";
        for (const ControlDevicePtr& controlDevice : controlDevices.values())
        {
            ARMARX_CHECK_EXPRESSION(controlDevice);
            ARMARX_DEBUG << "----" << controlDevice->getDeviceName();
            if (!controlDevice->hasLVL0Controller(ControlModes::EmergencyStop))
            {
                std::stringstream s;
                s << "ControlDevice " << controlDevice->getDeviceName()
                  << " has no LVL0Controller with ControlMode " << ControlModes::EmergencyStop
                  << " (to fix this, add a LVL0Controller with this ControlMode to the ControlDevice).\nAvailable controllers: "
                  <<  controlDevice->getControlModes();
                ARMARX_ERROR << "--------" << s.str();
                throw LogicError {s.str()};

            }
            if (!controlDevice->hasLVL0Controller(ControlModes::StopMovement))
            {
                std::stringstream s;
                s << "ControlDevice " << controlDevice->getDeviceName()
                  << " has no LVL0Controller with ControlMode \"" << ControlModes::StopMovement
                  << "\" (to fix this, add a LVL0Controller with this ControlMode to the ControlDevice) \nAvailable controllers: "
                  << controlDevice->getControlModes();
                ARMARX_ERROR << "--------" << s.str();
                throw LogicError {s.str()};
            }
        }
        ARMARX_DEBUG << "checking sensor devices:";
        for (const SensorDevicePtr& sensorDevice : sensorDevices.values())
        {
            ARMARX_CHECK_EXPRESSION(sensorDevice);
            ARMARX_DEBUG << "----" << sensorDevice->getDeviceName();
            if (!sensorDevice->getSensorValue())
            {
                std::stringstream s;
                s << "SensorDevice " << sensorDevice->getSensorValue() << " has null SensorValue";
                ARMARX_ERROR << "--------" << s.str();
                throw LogicError {s.str()};
            }
            ARMARX_CHECK_EXPRESSION(sensorDevice);
        }

        ARMARX_DEBUG << "initializing buffers:";
        {
            //ctrl buffers
            {
                ARMARX_DEBUG << "----initializing controller buffers";
                controllersActivated.reinitAllBuffers(LVL0AndLVL1Controllers {getControllerBufferSize()});
                LVL0AndLVL1Controllers ctrlinitReq(getControllerBufferSize());
                for (std::size_t i = 0 ; i < getControllerBufferSize(); ++i)
                {
                    ctrlinitReq.lvl0Controllers.at(i) = controlDevices.at(i)->rtGetLVL0StopMovementController();
                }
                controllersRequested.reinitAllBuffers(ctrlinitReq);
                controllersRequested.commitWrite();//to make sure the diry bit is activated (to trigger rt switch)
            }
            //sensor/control buffer
            {
                ARMARX_DEBUG << "----initializing SensorValue/ControlTartget buffer";
                SensorAndControl scinitRead;
                SensorAndControl scinitHidden;
                SensorAndControl scinitWrite;
                scinitRead.sensors.reserve(getNumberOfSensorDevices());
                scinitHidden.sensors.reserve(getNumberOfSensorDevices());
                scinitWrite.sensors.reserve(getNumberOfSensorDevices());
                ARMARX_DEBUG << "--------cloning SensorValues";
                for (ConstSensorDevicePtr s : sensorDevices.values())
                {
                    ARMARX_DEBUG << "------------of device " << s->getDeviceName() << "(type = " << s->getSensorValueType() << ")";
                    scinitRead.sensors.emplace_back(s->getSensorValue()->clone());
                    scinitHidden.sensors.emplace_back(s->getSensorValue()->clone());
                    scinitWrite.sensors.emplace_back(s->getSensorValue()->clone());
                }
                scinitRead.control.reserve(getControllerBufferSize());
                scinitHidden.control.reserve(getControllerBufferSize());
                scinitWrite.control.reserve(getControllerBufferSize());
                ARMARX_DEBUG << "--------cloning ControlTargets";
                for (ConstControlDevicePtr c : controlDevices.values())
                {
                    ARMARX_DEBUG << "------------of device " << c->getDeviceName();
                    scinitRead.control.emplace_back();
                    scinitHidden.control.emplace_back();
                    scinitWrite.control.emplace_back();
                    auto& scinitReadCD = scinitRead.control.back();
                    auto& scinitHiddenCD = scinitHidden.control.back();
                    auto& scinitWriteCD = scinitWrite.control.back();
                    scinitReadCD.reserve(c->getLVL0Controllers().size());
                    scinitHiddenCD.reserve(c->getLVL0Controllers().size());
                    scinitWriteCD.reserve(c->getLVL0Controllers().size());
                    for (const LVL0Controller* lvl0 : c->getLVL0Controllers())
                    {
                        ARMARX_DEBUG << "----------------mode " << lvl0->getControlMode() << "(type = " << lvl0->getControlTarget()->getControlTargetType() << ")";
                        scinitReadCD.emplace_back(lvl0->getControlTarget()->clone());
                        scinitHiddenCD.emplace_back(lvl0->getControlTarget()->clone());
                        scinitWriteCD.emplace_back(lvl0->getControlTarget()->clone());
                    }
                }
                sensorAndControl.reinitAllBuffers(std::move(scinitWrite), std::move(scinitHidden), std::move(scinitRead));
            }
        }
        ARMARX_DEBUG << "copying device ptrs to const ptr map";
        {
            for (const auto& dev : controlDevices.values())
            {
                controlDevicesConstPtr[dev->getDeviceName()] = dev;
            }
            for (const auto& dev : sensorDevices.values())
            {
                sensorDevicesConstPtr[dev->getDeviceName()] = dev;
            }
        }
        ARMARX_VERBOSE << "ControlDevices:\n" << controlDevices.keys();
        ARMARX_VERBOSE << "SensorDevices:\n" << sensorDevices.keys();
        ARMARX_INFO << "finishDeviceInitialization...done! " << (MicroNow() - beg).count() << " us";
    }

    void RobotUnit::initializeDefaultUnits()
    {
        auto beg = MicroNow();
        {
            auto guard = getGuard();
            throwIfStateIsNot(RobotUnitState::InitializingUnits, __FUNCTION__);
            ARMARX_INFO << "initializing default units";
            initializeKinematicUnit();
            initializePlatformUnit();
            initializeForceTorqueUnit();
            initializeInertialMeasurementUnit();
        }
        ARMARX_INFO << "initializing default units...done! " << (MicroNow() - beg).count() << " us";
    }

    void RobotUnit::initializeKinematicUnit()
    {
        using UnitT = KinematicSubUnit;
        using IfaceT = KinematicUnitInterface;

        auto guard = getGuard();
        throwIfStateIsNot(RobotUnitState::InitializingUnits, __FUNCTION__);
        //check if unit is already added
        if (getUnit(IfaceT::ice_staticId()))
        {
            return;
        }
        //add ctrl et al
        std::map<std::string, UnitT::ActuatorData> devs;
        for (const ControlDevicePtr& controlDevice : controlDevices.values())
        {
            const auto& controlDeviceName = controlDevice->getDeviceName();
            if (!controlDevice->hasTag(ControlDeviceTags::CreateNoDefaultController))
            {
                UnitT::ActuatorData ad;
                ad.name = controlDeviceName;
                ad.sensorDeviceIndex = sensorDevices.has(controlDeviceName) ? sensorDevices.index(controlDeviceName) : std::numeric_limits<std::size_t>::max();
                /// TODO use better kinematic unit controllers (posThroughVel + ramps)
#define initializeKinematicUnit_tmp_helper(mode, CtargT, ctrl)                                                          \
    {                                                                                                                   \
        const auto& controlMode = mode;                                                                                 \
        LVL1KinematicUnitPassThroughControllerPtr& pt = ctrl;                                                           \
        LVL0Controller* lvl0 = controlDevice->getLVL0Controller(controlMode);                                           \
        if (lvl0 && lvl0->getControlTarget()->isA<CtargT>())                                                            \
        {                                                                                                               \
            LVL1KinematicUnitPassThroughControllerConfigPtr config = new LVL1KinematicUnitPassThroughControllerConfig;  \
            config->controlMode=controlMode;                                                                            \
            config->deviceName=controlDeviceName;                                                                       \
            LVL1ControllerPtr lvl1 = createLVL1Controller(                                                                  \
                                     "LVL1KinematicUnitPassThroughController",                                          \
                                     "LVL1KU_PTCtrl_"+controlDeviceName+"_"+controlMode,                                \
                                     config,                                                                            \
                                     false           );                                                                 \
            pt = LVL1KinematicUnitPassThroughControllerPtr::dynamicCast(lvl1);                                          \
            ARMARX_CHECK_EXPRESSION(pt);                                                                                \
        }                                                                                                               \
    }
                initializeKinematicUnit_tmp_helper(ControlModes::Position1DoF, ControlTarget1DoFActuatorPosition, ad.ctrlPos)
                initializeKinematicUnit_tmp_helper(ControlModes::Velocity1DoF, ControlTarget1DoFActuatorVelocity, ad.ctrlVel)
                initializeKinematicUnit_tmp_helper(ControlModes::Torque1DoF  , ControlTarget1DoFActuatorTorque  , ad.ctrlTor)
#undef initializeKinematicUnit_tmp_helper


                if (ad.ctrlPos || ad.ctrlTor || ad.ctrlVel)
                {
                    devs[controlDeviceName] = std::move(ad);
                }
            }
        }
        ARMARX_CHECK_EXPRESSION_W_HINT(!devs.empty(), "no 1DoF ControlDevice found with matching SensorDevice");
        //add it
        const std::string configName = getProperty<std::string>("KinematicUnitName");
        const std::string& configDomain = "ArmarX";
        const std::string confPre = configDomain + "." + configName + ".";
        Ice::PropertiesPtr properties = getIceProperties()->clone();
        //properties->setProperty(confPre + "MinimumLoggingLevel", getProperty<std::string>("MinimumLoggingLevel").getValue());
        //fill properties
        properties->setProperty(confPre + "RobotNodeSetName"    , robotNodeSetName);
        properties->setProperty(confPre + "RobotFileName"       , robotFileName);
        properties->setProperty(confPre + "RobotFileNameProject", robotProjectName);
        IceInternal::Handle<UnitT> unit = Component::create<UnitT>(properties , configName, configDomain);
        //fill devices (sensor + controller)
        unit->setupData(getProperty<std::string>("RobotFileName").getValue(), robot->clone(robot->getName()), std::move(devs), this);
        //add
        addUnit(unit);
    }

    void RobotUnit::initializePlatformUnit()
    {
        using UnitT = PlatformSubUnit;
        using IfaceT = PlatformUnitInterface;

        auto guard = getGuard();
        throwIfStateIsNot(RobotUnitState::InitializingUnits, __FUNCTION__);
        //check if unit is already added
        if (getUnit(IfaceT::ice_staticId()))
        {
            return;
        }
        //is there a platform dev?
        if (getRobotPlatformName().empty())
        {
            ARMARX_INFO << "no platform unit created since no platform name was given";
            return;
        }
        if (!controlDevices.has(getRobotPlatformName()))
        {
            ARMARX_WARNING << "no platform unit created since the platform control device with name '" << getRobotPlatformName() << "' was not found";
            return;
        }
        const ControlDevicePtr& controlDevice = controlDevices.at(robotPlatformName);
        const SensorDevicePtr&  sensorDevice = sensorDevices.at(robotPlatformName);
        LVL0Controller* lvl0vel = controlDevice->getLVL0Controller(ControlModes::HolonomicPlatformVelocity);
        ARMARX_CHECK_EXPRESSION(lvl0vel);
        ARMARX_CHECK_EXPRESSION(sensorDevice->getSensorValue()->asA<SensorValueHolonomicPlatform>());
        //add it
        const std::string configName = getProperty<std::string>("PlatformUnitName");
        const std::string& configDomain = "ArmarX";
        const std::string confPre = configDomain + "." + configName + ".";
        Ice::PropertiesPtr properties = getIceProperties()->clone();
        //properties->setProperty(confPre + "MinimumLoggingLevel", getProperty<std::string>("MinimumLoggingLevel").getValue());
        //fill properties
        properties->setProperty(confPre + "PlatformName", robotPlatformName);
        IceInternal::Handle<UnitT> unit = Component::create<UnitT>(properties , configName, configDomain);
        //config
        LVL1HolonomicPlatformUnitVelocityPassThroughControllerConfigPtr config = new LVL1HolonomicPlatformUnitVelocityPassThroughControllerConfig;
        config->initialVelocityX = 0;
        config->initialVelocityY = 0;
        config->initialVelocityRotation = 0;
        config->platformName = robotPlatformName;
        auto ctrl = LVL1HolonomicPlatformUnitVelocityPassThroughControllerPtr::dynamicCast(
                        createLVL1Controller("LVL1HolonomicPlatformUnitVelocityPassThroughController", configName + "_VelPTContoller", config, false)
                    );
        ARMARX_CHECK_EXPRESSION(ctrl);
        unit->pt = ctrl;
        unit->platformSensorIndex = sensorDevices.index(robotPlatformName);
        //add
        addUnit(unit);
    }

    void RobotUnit::initializeForceTorqueUnit()
    {
        using UnitT = ForceTorqueSubUnit;
        using IfaceT = ForceTorqueUnitInterface;

        auto guard = getGuard();
        throwIfStateIsNot(RobotUnitState::InitializingUnits, __FUNCTION__);
        //check if unit is already added
        if (getUnit(IfaceT::ice_staticId()))
        {
            return;
        }
        //check if it is required
        auto sensors = getRobot()->getSensors<VirtualRobot::ForceTorqueSensor>();
        std::vector<ForceTorqueSubUnit::DeviceData> ftdevs;
        ftdevs.reserve(sensors.size());
        for (const auto& ft : sensors)
        {
            auto deviceName = ft->getName();
            if (sensorDevices.has(deviceName))
            {
                const SensorDevicePtr& sensorDevice = sensorDevices.at(deviceName);
                if (sensorDevice->getSensorValue()->isA<SensorValueForceTorque>())
                {
                    ForceTorqueSubUnit::DeviceData dat;
                    dat.sensorIndex = sensorDevices.index(deviceName);
                    dat.deviceName = std::move(deviceName);
                    dat.frame = ft->getRobotNode()->getName();
                    ftdevs.emplace_back(std::move(dat));
                }
                else
                {
                    ARMARX_WARNING << "ForceTorqueSensorDevice " << deviceName << " has no SensorValueForceTorque";
                }
            }
            else
            {
                ARMARX_WARNING << "ForceTorqueSensor with name " << deviceName << " has no corresponding sensor device";
            }
        }
        if (ftdevs.empty())
        {
            ARMARX_WARNING << "no force torque unit created since there are no force torque sensor devices";
            return;
        }
        //add it
        const std::string configName = getProperty<std::string>("ForceTorqueUnitName");
        const std::string& configDomain = "ArmarX";
        const std::string confPre = configDomain + "." + configName + ".";
        Ice::PropertiesPtr properties = getIceProperties()->clone();
        //properties->setProperty(confPre + "MinimumLoggingLevel", getProperty<std::string>("MinimumLoggingLevel").getValue());
        //fill properties
        properties->setProperty(confPre + "AgentName", robot->getName());
        properties->setProperty(confPre + "ForceTorqueTopicName", getProperty<std::string>("ForceTorqueTopicName").getValue());
        ARMARX_INFO << "creating unit " << configName << " using these properties: " << properties->getPropertiesForPrefix("");
        IceInternal::Handle<UnitT> unit = Component::create<UnitT>(properties , configName, configDomain);
        unit->devs = ftdevs;
        //add
        addUnit(unit);
    }

    void RobotUnit::initializeInertialMeasurementUnit()
    {
        using UnitT = InertialMeasurementSubUnit;
        using IfaceT = InertialMeasurementUnitInterface;

        auto guard = getGuard();
        throwIfStateIsNot(RobotUnitState::InitializingUnits, __FUNCTION__);
        //check if unit is already added
        if (getUnit(IfaceT::ice_staticId()))
        {
            return;
        }
        //check if it is required
        std::map<std::string, std::size_t> imudevs;
        for (auto i : getIndices(sensorDevices.keys()))
        {
            const SensorDevicePtr& sensorDevice = sensorDevices.at(i);
            if (sensorDevice->getSensorValue()->isA<SensorValueIMU>())
            {
                imudevs[sensorDevice->getDeviceName()] = i;
            }
        }
        if (imudevs.empty())
        {
            ARMARX_IMPORTANT << "no inertial measurement unit created since there are no imu sensor devices";
            return;
        }
        //add it
        const std::string configName = getProperty<std::string>("InertialMeasurementUnitName");
        const std::string& configDomain = "ArmarX";
        const std::string confPre = configDomain + "." + configName + ".";
        Ice::PropertiesPtr properties = getIceProperties()->clone();
        //properties->setProperty(confPre + "MinimumLoggingLevel", getProperty<std::string>("MinimumLoggingLevel").getValue());
        //fill properties
        properties->setProperty(confPre + "IMUTopicName", getProperty<std::string>("IMUTopicName").getValue());
        IceInternal::Handle<UnitT> unit = Component::create<UnitT>(properties , configName, configDomain);
        unit->devs = imudevs;
        //add
        addUnit(unit);
    }

    void RobotUnit::addUnit(ManagedIceObjectPtr unit)
    {
        auto guard = getGuard();
        throwIfStateIsNot(RobotUnitState::InitializingUnits, __FUNCTION__);
        getArmarXManager()->addObjectAsync(unit, "", true, false);
        units.emplace_back(std::move(unit));
    }

    void RobotUnit::finishUnitInitialization()
    {

        auto beg = MicroNow();
        auto guard = getGuard();
        throwIfStateIsNot(RobotUnitState::InitializingUnits, __FUNCTION__);
        state = RobotUnitState::WaitingForRTThreadInitialization;
        //start publisher
        int pubtimestep = getProperty<std::size_t>("PublishPeriodMs");
        if (!pubtimestep)
        {
            pubtimestep = 1;
        }
        publishNewSensorDataTime = TimeUtil::GetTime();
        publisherTask = new PublisherTaskT([&] {publish();}, pubtimestep, false, getName() + "_PublisherTask");
        ARMARX_INFO << "starting publisher with timestep " << pubtimestep;
        publisherTask->start();
        ARMARX_INFO << "finishUnitInitialization...done! " << (MicroNow() - beg).count() << " us";
    }

    void RobotUnit::initRTThread()
    {
        auto guard = getGuard();
        throwIfStateIsNot(RobotUnitState::WaitingForRTThreadInitialization, __FUNCTION__);
        rtThreadId = std::this_thread::get_id();
        state = RobotUnitState::Running;
    }

    bool RobotUnit::rtSwitchControllerSetup()
    {
        rtThreadTimingsSensorDevice->rtMarkRtSwitchControllerSetupStart();
        if (!rtRequestedControllersChanged())
        {
            rtThreadTimingsSensorDevice->rtMarkRtSwitchControllerSetupEnd();
            return false;
        }

        //handle lvl1
        {
            //"merge"
            std::size_t n = rtGetControllerBufferSize();
            std::size_t idxAct = 0;
            std::size_t idxReq = 0;
            for (std::size_t i = 0; i < 2 * n; ++i)
            {
                //skip nullptrs in act
                while (idxAct < n && !rtGetActivatedLVL1Controllers().at(idxAct))
                {
                    ++idxAct;
                }
                const LVL1ControllerPtr& req = idxReq < n ? rtGetRequestedLVL1Controllers().at(idxReq) : nullptr; //may be null
                const LVL1ControllerPtr& act = idxAct < n ? rtGetActivatedLVL1Controllers().at(idxAct) : nullptr; //may be null if it is the last
                std::size_t reqId = req ? req->rtGetId() : std::numeric_limits<std::size_t>::max();
                std::size_t actId = act ? act->rtGetId() : std::numeric_limits<std::size_t>::max();

                if (reqId < actId)
                {
                    //new ctrl
                    RtActivateLVL1Controller(req);
                    ++idxReq;
                }
                else if (reqId > actId)
                {
                    RtDeactivateLVL1Controller(act);
                    ++idxAct;
                }
                else //if(reqId == actId)
                {
                    //same ctrl or both null ctrl
                    ++idxReq;
                    ++idxAct;
                }
                if (idxAct >= n && !req)
                {
                    break;
                }
            }
            rtGetActivatedLVL1Controllers() = rtGetRequestedLVL1Controllers();
        }

        //handle lvl0
        {
            for (std::size_t i = 0; i < rtGetControllerBufferSize(); ++i)
            {
                auto& controlDevice = controlDevices.at(i);
                const auto requestedLVL0 = rtGetRequestedLVL0Controllers().at(i);
                controlDevice->rtSetActiveLVL0Controller(requestedLVL0);
                rtGetActivatedLVL0Controllers().at(i) = requestedLVL0;
            }
        }
        rtWriteActivatedControllers();
        rtThreadTimingsSensorDevice->rtMarkRtSwitchControllerSetupEnd();
        return true;
    }

    void RobotUnit::rtDeactivateAssignedLVL1Controller(std::size_t ctrlDevIndex, bool writeActivatedControlers)
    {
        std::size_t lvl1Index = std::numeric_limits<std::size_t>::max();
        for (std::size_t i = 0; i < rtGetControllerBufferSize(); ++i)
        {
            const LVL1ControllerPtr& lvl1 = rtGetActivatedLVL1Controllers().at(i);
            if (lvl1 && lvl1->rtUsesControlDevice(ctrlDevIndex))
            {
                //found corresponding lvl1
                lvl1Index = i;
                break;
            }
        }
        ARMARX_CHECK_EXPRESSION_W_HINT(
            lvl1Index < rtGetControllerBufferSize(),
            lvl1Index << " < " << rtGetControllerBufferSize() <<
            ": no lvl1 controller controls this device (name = "
            << controlDevices.at(ctrlDevIndex)->getDeviceName()
            << ", ControlMode = " << rtGetActivatedLVL0Controllers().at(ctrlDevIndex)->getControlMode() << ")!" << "\n"
            << VAROUT(rtGetActivatedLVL0Controllers()) << "\n"
            << VAROUT(rtGetActivatedLVL1Controllers())
        );
        const LVL1ControllerPtr& lvl1 = rtGetActivatedLVL1Controllers().at(lvl1Index);
        RtDeactivateLVL1ControllerBecauseOfError(lvl1);
        for (auto ctrlDevIdx : lvl1->rtGetControlDeviceUsedIndices())
        {
            const ControlDevicePtr& controlDevice = controlDevices.at(ctrlDevIdx);
            LVL0Controller* es = controlDevice->rtGetLVL0EmergencyStopController();
            controlDevice->rtSetActiveLVL0Controller(es);
            rtGetActivatedLVL0Controllers().at(ctrlDevIdx) = es;
        }
        rtGetActivatedLVL1Controllers().at(lvl1Index) = nullptr;

        if (writeActivatedControlers)
        {
            ARMARX_IMPORTANT << "SWITCH CTRLS deac";
            rtWriteActivatedControllers();
        }
    }

    void RobotUnit::rtRunLVL1Controllers(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration)
    {
        rtThreadTimingsSensorDevice->rtMarkRtRunLVL1ControllersStart();
        for (LVL1Controller* lvl1 : rtGetActivatedLVL1Controllers())
        {
            try
            {
                if (lvl1)
                {
                    lvl1->rtSwapBufferAndRun(sensorValuesTimestamp, timeSinceLastIteration);
                }
            }
            catch (...)
            {
                ARMARX_INFO << "Lvl1 Controller " << lvl1->getInstanceName() << " threw an exception and is now deactivated: " << GetHandledExceptionString();
                RtSetLVL1ControllerActive(lvl1, 0, true);
            }
        }
        rtThreadTimingsSensorDevice->rtMarkRtRunLVL1ControllersEnd();
    }

    void RobotUnit::rtRunLVL0Controllers(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration)
    {
        rtThreadTimingsSensorDevice->rtMarkRtRunLVL0ControllersStart();
        for (const ControlDevicePtr& device : this->controlDevices.values())
        {
            device->rtRun(sensorValuesTimestamp, timeSinceLastIteration);
        }
        rtThreadTimingsSensorDevice->rtMarkRtRunLVL0ControllersEnd();
    }

    void RobotUnit::rtReadSensorDeviceValues(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration)
    {
        rtThreadTimingsSensorDevice->rtMarkRtReadSensorDeviceValuesStart();
        for (const SensorDevicePtr& device : this->sensorDevices.values())
        {
            device->rtReadSensorValues(sensorValuesTimestamp, timeSinceLastIteration);
        }
        rtThreadTimingsSensorDevice->rtMarkRtReadSensorDeviceValuesEnd();
    }

    void RobotUnit::rtHandleInvalidTargets()
    {
        rtThreadTimingsSensorDevice->rtMarkRtHandleInvalidTargetsStart();
        bool oneIsInvalid = 0;
        for (std::size_t i = 0; i < rtGetControllerBufferSize(); ++i)
        {
            if (!rtGetActivatedLVL0Controllers().at(i)->rtIsTargetVaild())
            {
                rtDeactivateAssignedLVL1Controller(i, false);
                oneIsInvalid = true;
            }
        }
        if (oneIsInvalid)
        {
            ARMARX_VERBOSE << "ONE IS INVALID";
            rtWriteActivatedControllers();
        }
        rtThreadTimingsSensorDevice->rtMarkRtHandleInvalidTargetsEnd();
    }

    void RobotUnit::rtResetAllTargets()
    {
        rtThreadTimingsSensorDevice->rtMarkRtResetAllTargetsStart();
        for (const ControlDevicePtr& controlDev : controlDevices.values())
        {
            controlDev->rtGetActiveLVL0Controller()->rtResetTarget();
        }
        rtThreadTimingsSensorDevice->rtMarkRtResetAllTargetsEnd();
    }

    ConstSensorDevicePtr RobotUnit::getSensorDevice(const std::string& sensorDeviceName) const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return sensorDevices.at(sensorDeviceName, SensorDevice::NullPtr);
    }

    ConstControlDevicePtr RobotUnit::getControlDevice(const std::string& deviceName)
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return controlDevices.at(deviceName, ControlDevice::NullPtr);
    }

    ControlTargetBase* RobotUnit::getControlTarget(const std::string& deviceName, const std::string& controlMode)
    {
        static const std::string fname = __FUNCTION__;
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        if (std::this_thread::get_id() != initLVL1ThreadId)
        {
            throw std::logic_error
            {
                fname + ": Must be only called during instantiation of a LVL1 controller from the instantiated controler " +
                "(Don't try to call this function outside of a LVL1Controller's  constructor)"
            };
        }
        if (initLVL1UsedControlTargets.find(deviceName) != initLVL1UsedControlTargets.end())
        {
            throw std::logic_error
            {
                fname + ": Must not request two ControlTargets for the same device. (got = " +
                initLVL1UsedControlTargets.at(deviceName) + ", requested " + controlMode + ") " +
                "(You can only have a ControlDevice in one ControlMode. " +
                "To check all available ControlModes for this device, get the ControlDevice via getControlDevice(" + deviceName + ") and querry it)"
            };
        }
        // we should allow these modes for debugging reasons + there is no real reason to forbid them
        //if (controlMode == ControlModes::EmergencyStop)
        //{
        //    throw std::invalid_argument {"RobotUnit::getControlTarget: Can't get target for the control mode " + ControlModes::EmergencyStop};
        //}
        //if (controlMode == ControlModes::StopMovement)
        //{
        //    throw std::invalid_argument {"RobotUnit::getControlTarget: Can't get target for the control mode " + ControlModes::StopMovement};
        //}
        ControlDevicePtr controlDevice = controlDevices.at(deviceName, ControlDevice::NullPtr);
        if (!controlDevice)
        {
            return nullptr;
        }
        if (!controlDevice->hasLVL0Controller(controlMode))
        {
            return nullptr;
        }
        //there is a device and a target as was requested:
        initLVL1UsedControlTargets[deviceName] = controlMode;
        return controlDevice->getLVL0Controller(controlMode)->getControlTarget();
    }

    LVL1ControllerInterfacePrx RobotUnit::createLVL1Controller(const std::string& className, const std::string& instanceName, const LVL1ControllerConfigPtr& config, const Ice::Current&)
    {
        //no lock required
        return LVL1ControllerInterfacePrx::uncheckedCast(createLVL1Controller(className, instanceName, config, true)->getProxy(-1, true));
    }

    LVL1ControllerInterfacePrx RobotUnit::createLVL1ControllerFromVariantConfig(const std::string& className, const std::string& instanceName, const StringVariantBaseMap& variants, const Ice::Current&)
    {
        //no lock required
        checkLVL1ControllerClassName(className);
        if (!LVL1ControllerRegistry::get(className)->hasRemoteConfiguration())
        {
            std::stringstream ss;
            ss << "Requested controller class '" << className << "' allows no remote configuration" << LVL1ControllerRegistry::getKeys()
               << " (Implement 'static WidgetDescription::WidgetPtr " << className << "::GenerateConfigDescription()'"
               << " and 'static LVL1ControllerConfigPtr " << className << "::GenerateConfigFromVariants(const StringVariantBaseMap&)' to allow remote configuration";
            ARMARX_ERROR << ss.str();
            throw InvalidArgumentException {ss.str()};
        }
        return createLVL1Controller(className, instanceName, LVL1ControllerRegistry::get(className)->GenerateConfigFromVariants(variants), GlobalIceCurrent/*to select ice overload*/);
    }

    LVL1ControllerPtr RobotUnit::createLVL1Controller(const std::string& className, const std::string& instanceName, const LVL1ControllerConfigPtr& config, bool deletable)
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        if (instanceName.empty())
        {
            ARMARX_ERROR << "The instance name is empty! (give a unique name)";
            throw InvalidArgumentException {"The instance name is empty! (give a unique name)"};
        }
        //check if we would be able to create the class
        checkLVL1ControllerClassName(className);
        auto& factory = LVL1ControllerRegistry::get(className);

        //check if the instance name is already in use
        if (lvl1Controllers.count(instanceName))
        {
            std::stringstream ss;
            ss << "There already is a controller instance with the name '" << instanceName << "'. Use a different instance name instead."
               << " Other used instance names are " << getLVL1ControllerNames();
            ARMARX_ERROR << ss.str();
            throw InvalidArgumentException {ss.str()};
        }
        //create the controller
        ARMARX_CHECK_EXPRESSION(initLVL1UsedControlTargets.empty());//in case someone broke things
        initLVL1ThreadId = std::this_thread::get_id();
        ARMARX_ON_SCOPE_EXIT
        {
            initLVL1UsedControlTargets.clear();
            initLVL1ThreadId = std::thread::id{};
        };
        LVL1ControllerPtr lvl1 = factory->create(LVL1ControllerDescriptionProviderInterfacePtr::dynamicCast(RobotUnitPtr {this}), config, robot);

        std::vector<char> ctrlDeviceUsedBitmap(getNumberOfControlDevices(), false);
        std::vector<std::size_t> ctrlDeviceUsedIndices;
        ctrlDeviceUsedIndices.reserve(getNumberOfControlDevices());
        for (const auto& cd2targ : initLVL1UsedControlTargets)
        {
            auto idx = controlDevices.index(cd2targ.first);
            ctrlDeviceUsedBitmap.at(idx) = true;
            ctrlDeviceUsedIndices.emplace_back(idx);
        }
        InitLVL1Controler(
            lvl1,
            this,
            lvl1ControllerNextId++,
            initLVL1UsedControlTargets,
            std::move(ctrlDeviceUsedBitmap),
            std::move(ctrlDeviceUsedIndices),
            deletable);
        getArmarXManager()->addObjectAsync(lvl1, instanceName, false, false);
        lvl1Controllers[instanceName] = lvl1;
        listenerPrx->lvl1ControllerCreated(instanceName);
        return lvl1;
    }

    void RobotUnit::setActivateControllersRequest(std::set<LVL1ControllerPtr, LVL1Controller::IdComp>&& ctrls)
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        //erase nullptr
        ctrls.erase(nullptr);
        ARMARX_CHECK_EXPRESSION(ctrls.size() <= getControllerBufferSize());
        if (std::set<LVL1ControllerPtr, LVL1Controller::IdComp> {getRequestedLVL1Controllers().begin(), getRequestedLVL1Controllers().end()} == ctrls)
        {
            //redirty the flag to swap in the same set again
            controllersRequested.commitWrite();
            return;
        }
        LVL0AndLVL1Controllers request {getControllerBufferSize()};
        std::size_t idx = 0;
        for (const LVL1ControllerPtr& lvl1 : ctrls)
        {
            request.lvl1Controllers.at(idx++) = lvl1.get();
            //add lvl0 for this ctrl
            for (const auto& cd2cm : lvl1->getControlDeviceUsedControlModeMap())
            {
                std::size_t lvl0idx = controlDevices.index(cd2cm.first);
                if (request.lvl0Controllers.at(lvl0idx))
                {
                    std::stringstream ss;
                    ss << "RobotUnit:setActivateControllersRequest controllers to activate are in conflict!\ncontrollers:";
                    for (const auto& ctrl : ctrls)
                    {
                        ss << "\n" << ctrl->getInstanceName();
                    }
                    ARMARX_ERROR << ss.str();
                    throw InvalidArgumentException {ss.str()};
                }
                request.lvl0Controllers.at(lvl0idx) = controlDevices.at(lvl0idx)->getLVL0Controller(cd2cm.second);
            }
        }
        for (auto i : getIndices(request.lvl0Controllers))
        {
            LVL0Controller*& lvl0 = request.lvl0Controllers.at(i);
            if (!lvl0)
            {
                LVL0Controller* lvl0Old = getRequestedLVL0Controllers().at(i);
                if (lvl0Old == controlDevices.at(i)->getLVL0StopMovementController())
                {
                    //don't replace this controller with emergency stop
                    lvl0 = lvl0Old;
                }
                else
                {
                    //no one controls this device -> emergency stop
                    lvl0 = controlDevices.at(i)->getLVL0EmergencyStopController();
                }
            }
        }
        writeRequestedControllers(std::move(request));
        ARMARX_INFO << "requested controllers:\n" << getRequestedLVL1ControllerNames();
    }

    void RobotUnit::deleteLVL1Controller(const std::string& name, const Ice::Current&)
    {
        deleteLVL1Controllers({name});
    }

    void RobotUnit::deleteLVL1Controllers(const Ice::StringSeq& names, const Ice::Current&)
    {
        ARMARX_VERBOSE << "requesting controller deletion for:\n" << names;
        if (names.empty())
        {
            return;
        }
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        auto ctrlsToActVec = getLVL1ControllersNotNull(names); //also checks if these controllers exist
        ARMARX_DEBUG << "all controllers requested to delete exist" << std::flush;
        //check if all can be deleted
        for (const auto& lvl1 : ctrlsToActVec)
        {
            if (!lvl1->isDeletable())
            {
                throw LogicError {"The LVL1Controller '" + lvl1->getInstanceName() + "' can't be deleted since this operation is not allowed for this controller! (no LVL1Controller was deleted)"};
            }
        }
        for (const auto& lvl1 : ctrlsToActVec)
        {
            if (lvl1->isControllerActive() || lvl1->isControllerRequested())
            {
                throw LogicError {"The LVL1Controller '" + lvl1->getInstanceName() + "' can't be deleted since it is active or requested! (no LVL1Controller was deleted)"};
            }
        }
        for (const auto& lvl1 : ctrlsToActVec)
        {
            const auto name = lvl1->getInstanceName();
            //deletion is done in a different thread since this call may be done by the controller (protection against use after free)
            lvl1ControllersToBeDeleted[name] = std::move(lvl1);
            lvl1Controllers.erase(name);
            ARMARX_VERBOSE << "added LVL1Controller '" << name << "' to be deleted";
        }
    }

    void RobotUnit::activateLVL1Controller(const std::string& name, const Ice::Current&)
    {
        activateLVL1Controllers({name});
    }

    void RobotUnit::activateLVL1Controllers(const Ice::StringSeq& names, const Ice::Current&)
    {
        ARMARX_VERBOSE << "requesting controller activation for:\n" << names;
        if (names.empty())
        {
            return;
        }
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        auto ctrlsToActVec = getLVL1ControllersNotNull(names); //also checks if these controllers exist
        ARMARX_DEBUG << "all requested controllers exist" << std::flush;
        std::set<LVL1ControllerPtr, LVL1Controller::IdComp> ctrlsToAct {ctrlsToActVec.begin(), ctrlsToActVec.end()};
        ARMARX_CHECK_EXPRESSION(!ctrlsToAct.count(nullptr));
        std::set<LVL1ControllerPtr, LVL1Controller::IdComp> ctrls {getRequestedLVL1Controllers().begin(), getRequestedLVL1Controllers().end()};
        ctrls.erase(nullptr);
        //check for conflict
        std::vector<char> inuse;
        //check requested controllers
        {
            auto r = LVL1Controller::AreNotInConflict(ctrlsToAct.begin(), ctrlsToAct.end());
            if (!r)
            {
                std::stringstream ss;
                ss << "RobotUnit::activateLVL1Controllers: requested controllers are in conflict!\ncontrollers:\n" << names;
                ARMARX_ERROR << ss.str();
                throw InvalidArgumentException {ss.str()};
            }
            inuse = std::move(r.get());
        }
        ARMARX_DEBUG << "all requested controllers are conflict free" << std::flush;
        ARMARX_DEBUG << "inuse field \n" << inuse;
        //add already active controllers if they are conflict free
        {
            if (ctrls.empty())
            {
                ARMARX_DEBUG << "no already requested LVL1Controllers";
            }
            for (const LVL1ControllerPtr& lvl1 : ctrls)
            {
                if (ctrlsToAct.count(lvl1))
                {
                    continue;
                }
                auto r = lvl1->isNotInConflictWith(inuse);
                if (r)
                {
                    ARMARX_DEBUG << "keeping  already requested LVL1Controller '" << lvl1->getInstanceName() << "' in   list of requested controllers";
                    ctrlsToAct.insert(lvl1);
                    inuse = std::move(r.get());
                }
                else
                {
                    ARMARX_INFO << "removing already requested LVL1Controller '" << lvl1->getInstanceName() << "' from list of requested controllers";
                }
            }
            ARMARX_DEBUG << "inuse field \n" << inuse;
        }
        setActivateControllersRequest(std::move(ctrlsToAct));
    }

    void RobotUnit::deactivateLVL1Controller(const std::string& name, const Ice::Current&)
    {
        deactivateLVL1Controllers(Ice::StringSeq {name});
    }

    void RobotUnit::deactivateLVL1Controllers(const Ice::StringSeq& names, const Ice::Current&)
    {
        ARMARX_VERBOSE << "requesting controller deactivation for:\n" << names;
        if (names.empty())
        {
            return;
        }
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        auto ctrlsDeacVec = getLVL1ControllersNotNull(names); //also checks if these controllers exist
        std::set<LVL1ControllerPtr, LVL1Controller::IdComp> ctrls {getRequestedLVL1Controllers().begin(), getRequestedLVL1Controllers().end()};
        for (const auto& lvl1ToDeactivate : ctrlsDeacVec)
        {
            ctrls.erase(lvl1ToDeactivate);
        }
        setActivateControllersRequest(std::move(ctrls));
    }

    void RobotUnit::switchLVL1ControllerSetup(const Ice::StringSeq& newSetup, const Ice::Current&)
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        auto ctrlsToActVec = getLVL1ControllersNotNull(newSetup); //also checks if these controllers exist
        setActivateControllersRequest({ctrlsToActVec.begin(), ctrlsToActVec.end()});
    }

    void RobotUnit::publish()
    {
        auto begPublish = Now();
        static const int spamdelay = 1;
        std::set<RobotUnitState> skipStates
        {
            RobotUnitState::PreComponentInitialization,
            RobotUnitState::InitializingDevices,
            RobotUnitState::InitializingUnits,
            RobotUnitState::WaitingForRTThreadInitialization,
            RobotUnitState::Exiting
        };
        if (skipStates.count(state))
        {
            ARMARX_DEBUG << deactivateSpam(spamdelay) << "skipping publishing in state " << state;
            return;
        }
        auto guard = getGuard();
        if (skipStates.count(state))
        {
            ARMARX_DEBUG << deactivateSpam(spamdelay) << "skipping publishing in state " << state;
            return;
        }
        throwIfStateIsNot(RobotUnitState::Running, __FUNCTION__);
        //get batch proxies
        auto debugObserverBatchPrx = debugObserverPrx->ice_batchOneway();
        auto listenerBatchPrx = listenerPrx->ice_batchOneway();
        //delete lvl1 queued for deletion
        {
            for (auto& pair : lvl1ControllersToBeDeleted)
            {
                listenerBatchPrx->lvl1ControllerDeleted(pair.first);
                getArmarXManager()->removeObjectNonBlocking(pair.second);
                ARMARX_VERBOSE << "deleted LVL1Controller " << pair.first;
            }
            lvl1ControllersToBeDeleted.clear();
        }
        //swap buffers in
        const bool haveActivatedControllersChanged = activatedControllersChanged();
        const bool haveSensorAndControlValsChanged = sensorAndControlBufferChanged();
        //setup datastructures
        const auto& sensorAndControlBuffer =  getSensorAndControlBuffer();
        const auto& activatedControllers   = getActivatedControllers();
        const auto& requestedControllers   = getRequestedControllers();

        const auto timestamp = sensorAndControlBuffer.sensorValuesTimestamp;

        const auto numSensorDevices  = getNumberOfSensorDevices();
        const auto numControlDevices = getNumberOfControlDevices();

        const bool publishDebugObserver = !(publishIterationCount % debugObserverSkipIterations);
        const bool debugObserverPublishControlTargetsThisIteration = debugObserverPublishControlTargets;
        const bool debugObserverPublishSensorValuesThisIteration = debugObserverPublishSensorValues;
        StringVariantBaseMap debugObserverMap;

        if (haveSensorAndControlValsChanged)
        {
            //update units
            debugObserverMap["publishTimings_UnitUpdate"] = new TimedVariant
            {
                ARMARX_TIMER(TimestampVariant)
                {
                    ARMARX_DEBUG << deactivateSpam(spamdelay) << "updating units with new sensor values";
                    publishNewSensorDataTime = TimeUtil::GetTime();
                    for (ManagedIceObjectPtr& u : units)
                    {
                        RobotUnitSubUnitPtr rsu = RobotUnitSubUnitPtr::dynamicCast(u);
                        if (rsu && rsu->getObjectScheduler() && rsu->getProxy())
                        {
                            debugObserverMap["publishTimings_UnitUpdate_" + rsu->getName()] = new TimedVariant
                            {
                                ARMARX_TIMER(TimestampVariant){rsu->update(sensorAndControlBuffer, activatedControllers);},
                                timestamp
                            };
                        }
                    }
                },
                timestamp
            };
            //publish sensor updates

            debugObserverMap["publishTimings_SensorUpdates"] = new TimedVariant
            {
                ARMARX_TIMER(TimestampVariant)
                {
                    for (std::size_t sensidx = 0 ; sensidx < numSensorDevices; ++sensidx)
                    {
                        const SensorDevice& sensDev = *(sensorDevices.at(sensidx));
                        const SensorValueBase& sensVal = *(sensorAndControlBuffer.sensors.at(sensidx));
                        auto variants = sensVal.toVariants(timestamp);

                        if (!variants.empty())
                        {
                            if (debugObserverPublishSensorValuesThisIteration && publishDebugObserver)
                            {
                                transformMapKeys(
                                    variants, debugObserverMap,
                                    detail::makeDebugObserverNameFixer(sensDev.getDeviceName() + "_" + sensVal.getSensorValueType(true) + "_")
                                );
                            }
                            SensorDeviceStatus status;
                            status.deviceName = sensDev.getDeviceName();
                            status.sensorValue = std::move(variants);
                            status.timestampUSec = MicroNow().count();
                            listenerBatchPrx->sensorDeviceStatusChanged(status);
                        }
                    }
                },
                timestamp
            };
        }
        else
        {
            const double sensSecondsAgo = (TimeUtil::GetTime() - publishNewSensorDataTime).toSecondsDouble();
            if (sensSecondsAgo > 1)
            {
                ARMARX_WARNING << deactivateSpam(spamdelay) << "RobotUnit::publish: Last sensor value update is " << sensSecondsAgo << " seconds ago!";
            }
        }
        if (haveSensorAndControlValsChanged || haveActivatedControllersChanged)
        {
            debugObserverMap["publishTimings_ControlUpdates"] = new TimedVariant
            {
                ARMARX_TIMER(TimestampVariant)
                {
                    for (std::size_t ctrlidx = 0 ; ctrlidx < numControlDevices; ++ctrlidx)
                    {
                        const ControlDevice& ctrlDev = *(controlDevices.at(ctrlidx));

                        StringToStringVariantBaseMapMap variants;
                        for (const auto& ctrlVal : sensorAndControlBuffer.control.at(ctrlidx))
                        {
                            const auto& controlMode = ctrlVal->getControlMode();
                            variants[controlMode] = ctrlVal->toVariants(timestamp);
                            if (
                                haveSensorAndControlValsChanged &&
                                !variants[controlMode].empty() &&
                                debugObserverPublishControlTargetsThisIteration  &&
                                publishDebugObserver
                            )
                            {
                                transformMapKeys(
                                    variants[controlMode], debugObserverMap,
                                    detail::makeDebugObserverNameFixer(ctrlDev.getDeviceName() + "_" + controlMode + "_")
                                );
                            }
                        }
                        ControlDeviceStatus status;
                        const auto activeLVL0 = activatedControllers.lvl0Controllers.at(ctrlidx);
                        status.activeControlMode = activeLVL0 ? activeLVL0->getControlMode() : std::string {"!!LVL0Controller is nullptr!!"};
                        status.deviceName = ctrlDev.getDeviceName();
                        ARMARX_CHECK_EXPRESSION(requestedControllers.lvl0Controllers.at(ctrlidx));
                        status.requestedControlMode = requestedControllers.lvl0Controllers.at(ctrlidx)->getControlMode();
                        status.controlTargetValues = std::move(variants);
                        status.timestampUSec = MicroNow().count();
                        listenerBatchPrx->controlDeviceStatusChanged(status);
                    }
                },
                timestamp
            };
        }
        //call publish hook + publish LVL1Controller changes
        debugObserverMap["publishTimings_LVL1Updates"] = new TimedVariant
        {
            ARMARX_TIMER(TimestampVariant)
            {
                for (const LVL1ControllerPtr& lvl1 : getMapValues(lvl1Controllers))
                {
                    debugObserverMap["publishTimings_LVL1Updates_" + lvl1->getInstanceName()] = new TimedVariant
                    {
                        ARMARX_TIMER(TimestampVariant)
                        {
                            //run some hook for active (used for visu)
                            PublishLVL1Controller(lvl1, sensorAndControlBuffer, debugDrawerPrx, debugObserverBatchPrx);
                            if (
                                GetLVL1ControllerStatusReportedActive(lvl1) != lvl1->isControllerActive() ||
                                GetLVL1ControllerStatusReportedRequested(lvl1) != lvl1->isControllerRequested()
                            )
                            {
                                UpdateLVL1ControllerStatusReported(lvl1);
                                listenerBatchPrx->lVl1ControllerStatusChanged(lvl1->getControllerStatus());
                            }
                        },
                        timestamp
                    };
                }
            },
            timestamp
        };

        //report new class names
        debugObserverMap["publishTimings_ClassNameUpdates"] = new TimedVariant
        {
            ARMARX_TIMER(TimestampVariant)
            {
                const auto classNames = LVL1ControllerRegistry::getKeys();
                if (lastReportedClasses.size() != classNames.size())
                {
                    for (const auto& name : classNames)
                    {
                        if (!lastReportedClasses.count(name))
                        {
                            listenerBatchPrx->lvl1ControllerClassAdded(name);
                            lastReportedClasses.emplace(name);
                        }
                    }
                }
            },
            timestamp
        };
        debugObserverMap["publishTimings_RobotUnitListenerFlush"] = new TimedVariant {ARMARX_TIMER(TimestampVariant){listenerBatchPrx->ice_flushBatchRequests();}, timestamp};

        if (publishDebugObserver)
        {
            debugObserverMap["publishTimings_LastDebugObserverFlush"] = new TimedVariant {TimestampVariant{lastDebugObserverFlush.count()}, timestamp};
            debugObserverMap["publishTimings_LastPublishLoop"] = new TimedVariant {TimestampVariant{lastPublishLoop.count()}, timestamp};
            debugObserverBatchPrx->setDebugChannel(getName(), debugObserverMap);
            lastDebugObserverFlush = ARMARX_TIMER()
            {
                debugObserverBatchPrx->ice_flushBatchRequests();
            };
        }
        else
        {
            lastDebugObserverFlush = std::chrono::microseconds {0};
        }

        //warn if there are unset lvl0 controllers
        {
            const auto& lvl0s = getActivatedLVL0Controllers();
            if (std::any_of(lvl0s.begin(), lvl0s.end(), [](const LVL0Controller * lvl0)
        {
            return !lvl0;
        }))
            {
                ARMARX_WARNING << "RobotUnit::publish: Some activated LVL0Controllers are reported to be nullptr! "
                               << "(did you forget to call rtSwitchControllerSetup()?)";
            }
        }
        ++publishIterationCount;
        lastPublishLoop = MicroToNow(begPublish);
    }

    void RobotUnit::removeAllUnits()
    {
        for (ManagedIceObjectPtr& unit : units)
        {
            checkRefCountIsOne(unit, unit->getName());
            getArmarXManager()->removeObjectBlocking(unit);
            checkRefCountIsOne(unit, unit->getName());
        }
        units.clear();
    }

    void RobotUnit::finishRunning()
    {
        if (state == RobotUnitState::Exiting)
        {
            ARMARX_ERROR << "RobotUnit::finishRunning called multiple times!";
            return;
        }
        auto guard = getGuard();
        throwIfStateIsNot(RobotUnitState::Running, __FUNCTION__);
        state = RobotUnitState::Exiting;
        //const ptr maps
        controlDevicesConstPtr.clear();
        sensorDevicesConstPtr.clear();
        //publisher
        {
            publisherTask->stop();
            auto tPublisherStopped = Now();
            while (publisherTask->isFunctionExecuting() || publisherTask->isRunning())
            {
                ARMARX_FATAL << deactivateSpam(0.1) << "PUBLISHER TASK IS RUNNING EVEN AFTER IT WAS STOPPED!";
            }
            checkRefCountIsOne(publisherTask, "publisherTask");
            publisherTask = nullptr;
            //since the drawer queues draw events and we want to clear the layers, we have to sleep here
            std::this_thread::sleep_until(tPublisherStopped + std::chrono::milliseconds {100});
        }

        //RT
        joinRTThread();
        //call deac publishing of all active lvl1
        for (const auto& pair : lvl1Controllers)
        {
            ARMARX_DEBUG << "forcing deactivation of publishing for " << pair.first;
            DeactivateLVL1ControllerPublishing(pair.second, debugDrawerPrx, debugObserverPrx);
        }
        //units
        removeAllUnits();
        //lvl1 queued for deletion (some could still be in the queue)
        {
            for (auto& n2lvl1 : lvl1ControllersToBeDeleted)
            {
                LVL1ControllerPtr& lvl1 = n2lvl1.second;
                getArmarXManager()->removeObjectBlocking(lvl1);
                checkRefCountIsOne(lvl1, lvl1->getName());
            }
            lvl1ControllersToBeDeleted.clear();
        }
        //lvl1
        {
            for (auto& n2lvl1 : lvl1Controllers)
            {
                LVL1ControllerPtr& lvl1 = n2lvl1.second;
                getArmarXManager()->removeObjectBlocking(lvl1);
                checkRefCountIsOne(lvl1, lvl1->getName());
            }
            lvl1Controllers.clear();
        }
        //clear devices
        sensorDevices.clear();
        controlDevices.clear();
        //robot
        if (robot.use_count() != 1)
        {
            ARMARX_WARNING << "The robot's smart pointer use count is " << robot.use_count();
        }
        robot.reset();
        ARMARX_INFO << "RobotUnit finished running";
    }

    bool RobotUnit::loadLibFromPath(const std::string& path, const Ice::Current&)
    {
        std::string absPath;
        if (ArmarXDataPath::getAbsolutePath(path, absPath))
        {
            return loadLibFromAbsolutePath(absPath);
        }
        else
        {
            ARMARX_ERROR << "Could not find library " + path;
            return false;
        }
    }

    bool RobotUnit::loadLibFromPackage(const std::string& package, const std::string& libname, const Ice::Current&)
    {
        CMakePackageFinder finder(package);
        if (!finder.packageFound())
        {
            ARMARX_ERROR << "Could not find package '" << package << "'";
            return false;
        }

        for (auto libDirPath : Split(finder.getLibraryPaths(), ";"))
        {
            boost::filesystem::path fullPath = libDirPath;
            fullPath /= "lib" + libname  + "." + DynamicLibrary::GetSharedLibraryFileExtension();
            if (!boost::filesystem::exists(fullPath))
            {
                fullPath = libDirPath;
                fullPath /= libname;
                if (!boost::filesystem::exists(fullPath))
                {
                    continue;
                }
            }
            if (loadLibFromAbsolutePath(fullPath.string()))
            {
                return true;
            }
        }
        ARMARX_ERROR << "Could not find library " <<  libname << " in package " << package;
        return false;
    }

    bool RobotUnit::loadLibFromAbsolutePath(const std::string& path)
    {
        auto guard = getGuard();
        if (loadedLibs.count(path))
        {
            return true;
        }
        DynamicLibraryPtr lib(new DynamicLibrary());
        try
        {
            lib->load(path);
        }
        catch (...)
        {
            handleExceptions();
            return false;
        }

        if (lib->isLibraryLoaded())
        {
            loadedLibs[path] = lib;
        }
        else
        {
            ARMARX_ERROR << "Could not load lib " + path + ": " + lib->getErrorMessage();
            return false;
        }
        return true;
    }

    Ice::StringSeq RobotUnit::getLVL1ControllerClassNames(const Ice::Current&) const
    {
        return LVL1ControllerRegistry::getKeys();
    }

    LVL1ControllerClassDescription RobotUnit::getLVL1ControllerClassDescription(const std::string& className, const Ice::Current&) const
    {
        while (state == RobotUnitState::PreComponentInitialization || state == RobotUnitState::InitializingDevices)
        {
            //this phase should only last short so busy waiting is ok
            std::this_thread::sleep_for(std::chrono::milliseconds(50));
        }
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        checkLVL1ControllerClassName(className);
        LVL1ControllerClassDescription data;
        data.className = className;
        if (LVL1ControllerRegistry::get(className)->hasRemoteConfiguration())
        {
            data.configDescription = LVL1ControllerRegistry::get(className)->GenerateConfigDescription(robot, controlDevicesConstPtr, sensorDevicesConstPtr);
        }
        return data;
    }

    LVL1ControllerClassDescriptionSeq RobotUnit::getLVL1ControllerClassDescriptions(const Ice::Current&) const
    {
        std::size_t tries = 200;
        while (state == RobotUnitState::PreComponentInitialization || state == RobotUnitState::InitializingDevices)
        {
            //this phase should only last short so busy waiting is ok
            std::this_thread::sleep_for(std::chrono::milliseconds(50));
            if (! --tries)
            {
                throw RuntimeError {"RobotUnit::getLVL1ControllerClassDescriptions: it took too long to for the unit to get in a valid state"};
            }
        }
        auto guard = getGuard();
        LVL1ControllerClassDescriptionSeq r;
        r.reserve(LVL1ControllerRegistry::getKeys().size());
        for (const auto& key : LVL1ControllerRegistry::getKeys())
        {
            r.emplace_back(getLVL1ControllerClassDescription(key));
        }
        return r;
    }

    Ice::StringSeq RobotUnit::getLVL1ControllerNames(const Ice::Current&) const
    {
        auto guard = getGuard();
        return getMapKeys(lvl1Controllers);
    }

    Ice::StringSeq RobotUnit::getRequestedLVL1ControllerNames(const Ice::Current&) const
    {
        auto guard = getGuard();
        return GetNonNullNames(getRequestedLVL1Controllers());
    }

    Ice::StringSeq RobotUnit::getActivatedLVL1ControllerNames(const Ice::Current&) const
    {
        auto guard = getGuard();
        return GetNonNullNames(getActivatedLVL1Controllers());
    }

    LVL1ControllerInterfacePrx RobotUnit::getLVL1Controller(const std::string& name, const Ice::Current&) const
    {
        LVL1ControllerPtr ctrl;
        {
            auto guard = getGuard();
            auto it = lvl1Controllers.find(name);
            if (it == lvl1Controllers.end())
            {
                return nullptr;
            }
            ctrl = it->second;
        }
        return LVL1ControllerInterfacePrx::uncheckedCast(ctrl->getProxy(-1, true));
    }

    LVL1ControllerStatus RobotUnit::getLVL1ControllerStatus(const std::string& name, const Ice::Current&) const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return getLVL1ControllerNotNull(name)->getControllerStatus();
    }

    LVL1ControllerStatusSeq RobotUnit::getLVL1ControllerStatuses(const Ice::Current&) const
    {
        auto guard = getGuard();
        if (!areDevicesReady())
        {
            return {};
        }
        LVL1ControllerStatusSeq r;
        r.reserve(lvl1Controllers.size());
        for (const auto& lvl1 : lvl1Controllers)
        {
            r.emplace_back(lvl1.second->getControllerStatus());
        }
        return r;
    }

    LVL1ControllerDescription RobotUnit::getLVL1ControllerDescription(const std::string& name, const Ice::Current&) const
    {
        LVL1ControllerPtr lvl1;
        {
            auto guard = getGuard();
            throwIfDevicesNotReady(__FUNCTION__);
            lvl1 = getLVL1ControllerNotNull(name);
        }
        return lvl1->getControllerDescription();
    }

    LVL1ControllerDescriptionSeq RobotUnit::getLVL1ControllerDescriptions(const Ice::Current&) const
    {
        std::map<std::string, LVL1ControllerPtr> lvl1ControllersCopy;
        {
            auto guard = getGuard();
            if (!areDevicesReady())
            {
                return {};
            }
            lvl1ControllersCopy = lvl1Controllers;
        }
        LVL1ControllerDescriptionSeq r;
        r.reserve(lvl1ControllersCopy.size());
        for (const auto& lvl1 : lvl1ControllersCopy)
        {
            r.emplace_back(lvl1.second->getControllerDescription());
        }
        return r;
    }

    LVL1ControllerDescriptionWithStatus RobotUnit::getLVL1ControllerDescriptionWithStatus(const std::string& name, const Ice::Current&) const
    {
        LVL1ControllerPtr lvl1;
        {
            auto guard = getGuard();
            throwIfDevicesNotReady(__FUNCTION__);
            lvl1 = getLVL1ControllerNotNull(name);
        }
        return lvl1->getControllerDescriptionWithStatus();
    }

    LVL1ControllerDescriptionWithStatusSeq RobotUnit::getLVL1ControllerDescriptionsWithStatuses(const Ice::Current&) const
    {
        std::map<std::string, LVL1ControllerPtr> lvl1ControllersCopy;
        {
            auto guard = getGuard();
            if (!areDevicesReady())
            {
                return {};
            }
            lvl1ControllersCopy = lvl1Controllers;
        }
        LVL1ControllerDescriptionWithStatusSeq r;
        r.reserve(lvl1ControllersCopy.size());
        for (const auto& lvl1 : lvl1ControllersCopy)
        {
            r.emplace_back(lvl1.second->getControllerDescriptionWithStatus());
        }
        return r;
    }

    Ice::StringSeq RobotUnit::getControlDeviceNames(const Ice::Current&) const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return controlDevices.keys();
    }

    ControlDeviceDescription armarx::RobotUnit::RobotUnit::getControlDeviceDescription(std::size_t idx) const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        const ControlDevicePtr& controlDevice = controlDevices.at(idx);
        ControlDeviceDescription data;
        data.deviceName = controlDevice->getDeviceName();
        data.tags.assign(controlDevice->getTags().begin(), controlDevice->getTags().end());
        for (const auto& lvl0 : controlDevice->getLVL0Controllers())
        {
            data.contolModeToTargetType[lvl0->getControlMode()] = lvl0->getControlTarget()->getControlTargetType();
        }
        return data;
    }

    ControlDeviceDescription RobotUnit::getControlDeviceDescription(const std::string& name, const Ice::Current&) const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        if (!controlDevices.has(name))
        {
            std::stringstream ss;
            ss << "RobotUnit::getControlDeviceDescription: There is no ControlDevice '" << name
               << "'. There are these ControlDevices: " << controlDevices.keys();
            throw InvalidArgumentException {ss.str()};
        }
        return getControlDeviceDescription(controlDevices.index(name));
    }

    ControlDeviceDescriptionSeq RobotUnit::getControlDeviceDescriptions(const Ice::Current&) const
    {
        auto guard = getGuard();
        if (!areDevicesReady())
        {
            return {};
        }
        ControlDeviceDescriptionSeq r;
        r.reserve(getNumberOfControlDevices());
        for (auto idx : getIndices(controlDevices.values()))
        {
            r.emplace_back(getControlDeviceDescription(idx));
        }
        return r;
    }

    ControlDeviceStatus armarx::RobotUnit::RobotUnit::getControlDeviceStatus(std::size_t idx) const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        const ControlDevicePtr& controlDevice = controlDevices.at(idx);
        ControlDeviceStatus status;
        const auto activeLVL0 = getActivatedLVL0Controllers().at(idx);
        status.activeControlMode = activeLVL0 ? activeLVL0->getControlMode() : std::string {"!!LVL0Controller is nullptr!!"};
        status.deviceName = controlDevice->getDeviceName();
        ARMARX_CHECK_EXPRESSION(getRequestedLVL0Controllers().at(idx));
        status.requestedControlMode = getRequestedLVL0Controllers().at(idx)->getControlMode();
        for (const auto& targ : getSensorAndControlBuffer().control.at(idx))
        {
            status.controlTargetValues[targ->getControlMode()] = targ->toVariants(getSensorAndControlBuffer().sensorValuesTimestamp);
        }
        status.timestampUSec = MicroNow().count();
        return status;
    }

    ControlDeviceStatus RobotUnit::getControlDeviceStatus(const std::string& name, const Ice::Current&) const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        if (!controlDevices.has(name))
        {
            std::stringstream ss;
            ss << "RobotUnit::getControlDeviceStatus: There is no ControlDevice '" << name
               << "'. There are these ControlDevices: " << controlDevices.keys();
            throw InvalidArgumentException {ss.str()};
        }
        return getControlDeviceStatus(controlDevices.index(name));
    }

    ControlDeviceStatusSeq RobotUnit::getControlDeviceStatuses(const Ice::Current&) const
    {
        auto guard = getGuard();
        if (!areDevicesReady())
        {
            return {};
        }
        ControlDeviceStatusSeq r;
        r.reserve(getNumberOfControlDevices());
        for (auto idx : getIndices(controlDevices.values()))
        {
            r.emplace_back(getControlDeviceStatus(idx));
        }
        return r;
    }

    Ice::StringSeq RobotUnit::getSensorDeviceNames(const Ice::Current&) const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return sensorDevices.keys();
    }

    SensorDeviceDescription armarx::RobotUnit::RobotUnit::getSensorDeviceDescription(std::size_t idx) const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        const SensorDevicePtr& sensorDevice = sensorDevices.at(idx);
        SensorDeviceDescription data;
        data.deviceName = sensorDevice->getDeviceName();
        data.tags.assign(sensorDevice->getTags().begin(), sensorDevice->getTags().end());
        data.sensorValueType = sensorDevice->getSensorValueType();
        return data;
    }

    SensorDeviceDescription RobotUnit::getSensorDeviceDescription(const std::string& name, const Ice::Current&) const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        if (!sensorDevices.has(name))
        {
            std::stringstream ss;
            ss << "RobotUnit::getSensorDeviceDescription: There is no SensorDevice '" << name
               << "'. There are these SensorDevices: " << sensorDevices.keys();
            throw InvalidArgumentException {ss.str()};
        }
        return getSensorDeviceDescription(sensorDevices.index(name));
    }

    SensorDeviceDescriptionSeq RobotUnit::getSensorDeviceDescriptions(const Ice::Current&) const
    {
        auto guard = getGuard();
        if (!areDevicesReady())
        {
            return {};
        }
        SensorDeviceDescriptionSeq r;
        r.reserve(getNumberOfSensorDevices());
        for (auto idx : getIndices(sensorDevices.values()))
        {
            r.emplace_back(getSensorDeviceDescription(idx));
        }
        return r;
    }

    SensorDeviceStatus armarx::RobotUnit::RobotUnit::getSensorDeviceStatus(std::size_t idx) const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        const SensorDevicePtr& sensorDevice = sensorDevices.at(idx);
        SensorDeviceStatus status;
        status.deviceName = sensorDevice->getDeviceName();
        status.sensorValue = getSensorAndControlBuffer().sensors.at(idx)->toVariants(getSensorAndControlBuffer().sensorValuesTimestamp);
        status.timestampUSec = MicroNow().count();
        return status;
    }

    SensorDeviceStatus RobotUnit::getSensorDeviceStatus(const std::string& name, const Ice::Current&) const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        if (!sensorDevices.has(name))
        {
            std::stringstream ss;
            ss << "RobotUnit::getSensorDeviceStatus: There is no SensorDevice '" << name
               << "'. There are these SensorDevices: " << sensorDevices.keys();
            throw InvalidArgumentException {ss.str()};
        }
        return getSensorDeviceStatus(sensorDevices.index(name));
    }

    SensorDeviceStatusSeq RobotUnit::getSensorDeviceStatuses(const Ice::Current&) const
    {
        auto guard = getGuard();
        if (!areDevicesReady())
        {
            return {};
        }
        SensorDeviceStatusSeq r;
        r.reserve(getNumberOfSensorDevices());
        for (auto idx : getIndices(sensorDevices.values()))
        {
            r.emplace_back(getSensorDeviceStatus(idx));
        }
        return r;
    }

    std::vector<armarx::LVL1ControllerPtr> RobotUnit::getLVL1ControllersNotNull(const std::vector<std::string>& names) const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        std::vector<LVL1ControllerPtr> ctrl;
        ctrl.reserve(names.size());
        for (const auto& name : names)
        {
            auto it = lvl1Controllers.find(name);
            if (it == lvl1Controllers.end())
            {
                std::stringstream ss;
                ss << "RobotUnit: there is no LVL1Controller with name '" << name
                   << "'. Existing LVL1Controllers are: " << getLVL1ControllerNames();
                throw InvalidArgumentException {ss.str()};
            }
            if (!it->second)
            {
                std::stringstream ss;
                ss << "RobotUnit: The LVL1Controller with name '" << name
                   << "'. Is a nullptr! This should never be the case (invariant)!  \nMap:\n" << lvl1Controllers;
                ARMARX_FATAL << ss.str();
                throw InvalidArgumentException {ss.str()};
            }
            ctrl.emplace_back(it->second);
        }
        return ctrl;
    }
    LVL1ControllerPtr RobotUnit::getLVL1ControllerNotNull(const std::string& name) const
    {
        return getLVL1ControllersNotNull(std::vector<std::string> {name}).at(0);
    }

    StringLVL1ControllerPrxDictionary RobotUnit::getAllLVL1Controllers(const Ice::Current&) const
    {
        std::map<std::string, LVL1ControllerPtr> lvl1ControllersCopy;
        {
            auto guard = getGuard();
            //copy to keep lock retention time low
            lvl1ControllersCopy = lvl1Controllers;
        }
        StringLVL1ControllerPrxDictionary result;
        for (const auto& pair : lvl1ControllersCopy)
        {
            result[pair.first] = LVL1ControllerInterfacePrx::uncheckedCast(pair.second->getProxy(-1, true));
        }
        return result;
    }

    Ice::ObjectProxySeq RobotUnit::getUnits(const Ice::Current&) const
    {
        std::vector<ManagedIceObjectPtr> unitsCopy;
        {
            auto guard = getGuard();
            //copy to keep lock retention time low
            unitsCopy = units;
        }
        Ice::ObjectProxySeq r;
        r.reserve(unitsCopy.size());
        for (const ManagedIceObjectPtr& unit : unitsCopy)
        {
            r.emplace_back(unit->getProxy(-1, true));
        }
        return r;
    }

    const ManagedIceObjectPtr& RobotUnit::getUnit(const std::string& staticIceId) const
    {
        auto guard = getGuard();
        for (const ManagedIceObjectPtr& unit : units)
        {
            if (unit->ice_isA(staticIceId))
            {
                return unit;
            }
        }
        return ManagedIceObject::NullPtr;
    }

    Ice::ObjectPrx RobotUnit::getUnit(const std::string& staticIceId, const Ice::Current&) const
    {
        //no lock required
        ManagedIceObjectPtr unit = getUnit(staticIceId);
        if (unit)
        {
            return unit->getProxy(-1, true);
        }
        return nullptr;
    }

    KinematicUnitInterfacePrx RobotUnit::getKinematicUnit(const Ice::Current&) const
    {
        //no lock required
        return KinematicUnitInterfacePrx::uncheckedCast(getUnit(KinematicUnitInterface::ice_staticId(), GlobalIceCurrent));
    }

    ForceTorqueUnitInterfacePrx RobotUnit::getForceTorqueUnit(const Ice::Current&) const
    {
        //no lock required
        return ForceTorqueUnitInterfacePrx::uncheckedCast(getUnit(ForceTorqueUnitInterface::ice_staticId(), GlobalIceCurrent));
    }

    InertialMeasurementUnitInterfacePrx RobotUnit::getInertialMeasurementUnit(const Ice::Current&) const
    {
        //no lock required
        return InertialMeasurementUnitInterfacePrx::uncheckedCast(getUnit(InertialMeasurementUnitInterface::ice_staticId(), GlobalIceCurrent));
    }

    PlatformUnitInterfacePrx RobotUnit::getPlatformUnit(const Ice::Current&) const
    {
        //no lock required
        return PlatformUnitInterfacePrx::uncheckedCast(getUnit(PlatformUnitInterface::ice_staticId(), GlobalIceCurrent));
    }

    const VirtualRobot::RobotPtr& RobotUnit::getRobot() const
    {
        auto guard = getGuard();
        throwIfStateIsNot({RobotUnitState::InitializingDevices, RobotUnitState::InitializingUnits, RobotUnitState::WaitingForRTThreadInitialization, RobotUnitState::Running}, __FUNCTION__);
        ARMARX_CHECK_EXPRESSION(robot);
        return robot;
    }

    std::string RobotUnit::getRobotName() const
    {
        auto guard = getGuard();
        throwIfStateIsNot({RobotUnitState::InitializingDevices, RobotUnitState::InitializingUnits, RobotUnitState::WaitingForRTThreadInitialization, RobotUnitState::Running}, __FUNCTION__);
        return getRobot()->getName();
    }

    std::size_t RobotUnit::getNumberOfControlDevices() const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return controlDevices.size();
    }

    std::size_t RobotUnit::getNumberOfSensorDevices() const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return sensorDevices.size();
    }

    const DebugDrawerInterfacePrx& RobotUnit::getDebugDrawerProxy() const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return debugDrawerPrx;
    }

    std::string RobotUnit::getListenerTopicName(const Ice::Current&) const
    {
        auto guard = getGuard();
        return robotUnitListenerTopicName;
    }

    const RobotUnitListenerPrx& RobotUnit::getListenerProxy() const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return listenerPrx;
    }

    PropertyDefinitionsPtr RobotUnit::createPropertyDefinitions()
    {
        return PropertyDefinitionsPtr(new RobotUnitPropertyDefinitions(getConfigIdentifier()));
    }

    void RobotUnit::icePropertiesUpdated(const std::set<std::string>& changedProperties)
    {

        if (changedProperties.count("DebugObserverPublishSensorValues"))
        {
            debugObserverPublishSensorValues = getProperty<bool>("DebugObserverPublishSensorValues").getValue();
        }
        if (changedProperties.count("DebugObserverPublishControlTargets"))
        {
            debugObserverPublishControlTargets = getProperty<bool>("DebugObserverPublishControlTargets").getValue();
        }
        if (changedProperties.count("DebugObserverPrintEveryNthIterations"))
        {
            debugObserverSkipIterations = getProperty<std::uint64_t>("DebugObserverPrintEveryNthIterations").getValue();
        }
    }

    void RobotUnit::onInitComponent()
    {
        ARMARX_DEBUG << "RobotUnit::onInitComponent()";
        {
            auto guard = getGuard();
            getArmarXManager()->increaseSchedulers(static_cast<int>(getProperty<std::uint64_t>("AdditionalObjectSchedulerCount").getValue()));

            throwIfStateIsNot(RobotUnitState::PreComponentInitialization, __FUNCTION__);
            robotUnitListenerTopicName = getProperty<std::string>("RobotUnitListenerTopicName").getValue();
            offeringTopic(getListenerTopicName());
            offeringTopic(getProperty<std::string>("DebugDrawerUpdatesTopicName").getValue());

            debugObserverPublishSensorValues = getProperty<bool>("DebugObserverPublishSensorValues").getValue();
            debugObserverPublishControlTargets = getProperty<bool>("DebugObserverPublishControlTargets").getValue();

            //load robot
            {
                robotNodeSetName            = getProperty<std::string>("RobotNodeSetName").getValue();
                robotProjectName            = getProperty<std::string>("RobotFileNameProject").getValue();
                robotFileName               = getProperty<std::string>("RobotFileName").getValue();
                robotPlatformName           = getProperty<std::string>("PlatformName").getValue();
                debugObserverSkipIterations = getProperty<std::uint64_t>("DebugObserverPrintEveryNthIterations").getValue();
                if (debugObserverSkipIterations < 1)
                {
                    debugObserverSkipIterations = 1;
                }
                Ice::StringSeq includePaths;
                if (!robotProjectName.empty())
                {
                    CMakePackageFinder finder(robotProjectName);
                    Ice::StringSeq projectIncludePaths;
                    auto pathsString = finder.getDataDir();
                    boost::split(projectIncludePaths,
                                 pathsString,
                                 boost::is_any_of(";,"),
                                 boost::token_compress_on);
                    includePaths.insert(includePaths.end(), projectIncludePaths.begin(), projectIncludePaths.end());
                }
                if (!ArmarXDataPath::getAbsolutePath(robotFileName, robotFileName, includePaths))
                {
                    throw UserException("Could not find robot file " + robotFileName);
                }
                // read the robot
                try
                {
                    robot = VirtualRobot::RobotIO::loadRobot(robotFileName, VirtualRobot::RobotIO::eFull);
                }
                catch (VirtualRobot::VirtualRobotException& e)
                {
                    throw UserException(e.what());
                }
                ARMARX_INFO << "Loaded robot:"
                            << "\n\tProject      : " << robotProjectName
                            << "\n\tRobot file   : " << robotFileName
                            << "\n\tRobotNodeSet : " << robotNodeSetName
                            << "\n\tNodeNames    : " << robot->getRobotNodeSet(robotNodeSetName)->getNodeNames();
            }
            ARMARX_CHECK_EXPRESSION(robot);
            state = RobotUnitState::InitializingDevices;
        }
        ARMARX_DEBUG << "RobotUnit::onInitComponent(): onInitRobotUnit()";
        onInitRobotUnit();
        ARMARX_DEBUG << "RobotUnit::onInitComponent()...done!";
    }

    void RobotUnit::onConnectComponent()
    {
        listenerPrx = getTopic<RobotUnitListenerPrx>(getListenerTopicName());
        debugDrawerPrx = getTopic<DebugDrawerInterfacePrx>(getProperty<std::string>("DebugDrawerUpdatesTopicName").getValue());
        debugObserverPrx = getTopic<DebugObserverInterfacePrx>(getProperty<std::string>("DebugObserverTopicName").getValue());
        onConnectRobotUnit();
    }

    void RobotUnit::onDisconnectComponent()
    {
        onDisconnectRobotUnit();
    }

    void RobotUnit::onExitComponent()
    {
        onExitRobotUnit();
        finishRunning();
    }

    bool RobotUnit::rtRequestedControllersChanged() const
    {
        return controllersRequested.updateReadBuffer();
    }

    const LVL0AndLVL1Controllers& RobotUnit::rtGetRequestedControllers() const
    {
        return controllersRequested.getReadBuffer();
    }

    const std::vector<armarx::LVL0Controller*>& RobotUnit::rtGetRequestedLVL0Controllers() const
    {
        return rtGetRequestedControllers().lvl0Controllers;
    }

    const std::vector<armarx::LVL1Controller*>& RobotUnit::rtGetRequestedLVL1Controllers() const
    {
        return rtGetRequestedControllers().lvl1Controllers;
    }

    void RobotUnit::rtWriteActivatedControllers()
    {
        controllersActivated.commitWrite();
    }

    LVL0AndLVL1Controllers& RobotUnit::rtGetActivatedControllers()
    {
        return controllersActivated.getWriteBuffer();
    }

    std::vector<armarx::LVL0Controller*>& RobotUnit::rtGetActivatedLVL0Controllers()
    {
        return rtGetActivatedControllers().lvl0Controllers;
    }

    std::vector<armarx::LVL1Controller*>& RobotUnit::rtGetActivatedLVL1Controllers()
    {
        return rtGetActivatedControllers().lvl1Controllers;
    }

    void RobotUnit::rtUpdateSensorAndControlBuffer(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration)
    {
        rtThreadTimingsSensorDevice->rtMarkRtUpdateSensorAndControlBufferStart();
        SensorAndControl& sc = sensorAndControl.getWriteBuffer();
        sc.sensorValuesTimestamp = sensorValuesTimestamp;
        sc.timeSinceLastIteration = timeSinceLastIteration;
        for (std::size_t sensIdx = 0; sensIdx < rtGetNumberOfSensorDevices(); ++sensIdx)
        {
            sensorDevices.at(sensIdx)->getSensorValue()->copyTo(sc.sensors.at(sensIdx));
        }
        for (std::size_t ctrlIdx = 0; ctrlIdx < controlDevices.size(); ++ctrlIdx)
        {
            ControlDevice& controlDevice = *controlDevices.at(ctrlIdx);
            auto& tv = sc.control.at(ctrlIdx);
            for (std::size_t targIdx = 0; targIdx < controlDevice.rtGetLVL0Controllers().size(); ++targIdx)
            {
                LVL0Controller& lvl0 = *controlDevice.rtGetLVL0Controllers().at(targIdx);
                lvl0.getControlTarget()->copyTo(tv.at(targIdx));
            }
        }
        sensorAndControl.commitWrite();
        rtThreadTimingsSensorDevice->rtMarkRtUpdateSensorAndControlBufferEnd();
    }

    void RobotUnit::writeRequestedControllers(LVL0AndLVL1Controllers &&setOfControllers)
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        //check lvl1
        const auto& lvl1s = setOfControllers.lvl1Controllers;
        std::set<LVL1ControllerPtr> lvl1set {lvl1s.begin(), lvl1s.end()};
        lvl1set.erase(nullptr);
        //# lvl1
        const std::size_t nlvl1s = std::count_if(lvl1s.begin(), lvl1s.end(), [](const LVL1ControllerPtr & p)
        {
            return p;
        });

        ARMARX_CHECK_EXPRESSION(nlvl1s == lvl1set.size());
        ARMARX_CHECK_EXPRESSION(lvl1s.size() == getControllerBufferSize());
        //first nlvl1s not null
        ARMARX_CHECK_EXPRESSION(std::all_of(lvl1s.begin(), lvl1s.begin() + nlvl1s, [](LVL1Controller * p)
        {
            return p;
        }));
        //last getNumberOfControlDevices()-nlvl1s null
        ARMARX_CHECK_EXPRESSION(std::all_of(lvl1s.begin() + nlvl1s, lvl1s.end(), [](LVL1Controller * p)
        {
            return !p;
        }));
        //conflict free and sorted
        ARMARX_CHECK_EXPRESSION(std::is_sorted(lvl1s.begin(), lvl1s.end(), LVL1Controller::IdComp {}));
        ARMARX_CHECK_EXPRESSION(LVL1Controller::AreNotInConflict(lvl1s.begin(), lvl1s.begin() + nlvl1s));

        //check lvl0
        const auto& lvl0s = setOfControllers.lvl0Controllers;
        ARMARX_CHECK_EXPRESSION(lvl0s.size() == getControllerBufferSize());
        ARMARX_CHECK_EXPRESSION(std::all_of(lvl0s.begin(), lvl0s.end(), [](LVL0Controller * p)
        {
            return p;
        }));
        checkSetOfControllersToActivate(setOfControllers);
        //set requested state
        for (const auto& name2lvl1 : lvl1Controllers)
        {
            const LVL1ControllerPtr& lvl1 = name2lvl1.second;
            SetLVL1ControllerRequested(lvl1, lvl1set.count(lvl1));
        }

        controllersRequested.getWriteBuffer() = std::move(setOfControllers);
        controllersRequested.commitWrite();
    }

    const LVL0AndLVL1Controllers& RobotUnit::getRequestedControllers() const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return controllersRequested.getWriteBuffer();
    }

    const std::vector<armarx::LVL0Controller*>& RobotUnit::getRequestedLVL0Controllers() const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return getRequestedControllers().lvl0Controllers;
    }

    const std::vector<armarx::LVL1Controller*>& RobotUnit::getRequestedLVL1Controllers() const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return getRequestedControllers().lvl1Controllers;
    }

    bool RobotUnit::activatedControllersChanged() const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return controllersActivated.updateReadBuffer();
    }

    const LVL0AndLVL1Controllers& RobotUnit::getActivatedControllers() const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return controllersActivated.getReadBuffer();
    }

    const std::vector<armarx::LVL0Controller*>& RobotUnit::getActivatedLVL0Controllers() const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return getActivatedControllers().lvl0Controllers;
    }

    const std::vector<armarx::LVL1Controller*>& RobotUnit::getActivatedLVL1Controllers() const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return getActivatedControllers().lvl1Controllers;
    }

    bool RobotUnit::sensorAndControlBufferChanged() const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return sensorAndControl.updateReadBuffer();
    }

    const SensorAndControl& RobotUnit::getSensorAndControlBuffer() const
    {
        auto guard = getGuard();
        throwIfDevicesNotReady(__FUNCTION__);
        return sensorAndControl.getReadBuffer();
    }

    std::ostream& operator<<(std::ostream& o, RobotUnitState s)
    {
        o << std::to_string(s);
        return o;
    }
}
