/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::RobotUnit
 * @author     Raphael ( ufdrv at student dot kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#ifndef _ARMARX_UNIT_RobotAPI_RobotUnit_H
#define _ARMARX_UNIT_RobotAPI_RobotUnit_H

#include <thread>
#include <atomic>
#include <chrono>
#include <mutex>
#include <unordered_map>
#include <memory>

#include <ArmarXCore/core/util/OnScopeExit.h>
#include <ArmarXCore/core/Component.h>
#include <ArmarXCore/core/ArmarXManager.h>
#include <ArmarXCore/core/time/TimeUtil.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <ArmarXCore/core/util/algorithm.h>
#include <ArmarXCore/core/util/TripleBuffer.h>
#include <ArmarXCore/core/services/tasks/TaskUtil.h>
#include <ArmarXCore/observers/DebugObserver.h>
#include <RobotAPI/interface/units/RobotUnit/RobotUnitInterface.h>

#include "LVL0Controllers/LVL0Controller.h"
#include "LVL1Controllers/LVL1Controller.h"
#include "Devices/ControlDevice.h"
#include "Devices/SensorDevice.h"
#include "Devices/RTThreadTimingsSensorDevice.h"
#include "util.h"
#include "structs.h"

namespace armarx
{
    TYPEDEF_PTRS_HANDLE(RobotSubUnitUpdateInterface);
    class RobotSubUnitUpdateInterface: virtual public ManagedIceObject {};

    TYPEDEF_PTRS_HANDLE(RobotUnit);

    class DynamicLibrary;
    typedef boost::shared_ptr<DynamicLibrary> DynamicLibraryPtr;
    /**
     * @class RobotUnitPropertyDefinitions
     * @brief
     */
    class RobotUnitPropertyDefinitions:
        public armarx::ComponentPropertyDefinitions
    {
    public:
        RobotUnitPropertyDefinitions(std::string prefix):
            armarx::ComponentPropertyDefinitions(prefix)
        {
            defineOptionalProperty<std::uint64_t>("AdditionalObjectSchedulerCount", 10, "Number of ObjectSchedulers to be added");

            defineOptionalProperty<std::string>("RobotUnitListenerTopicName", "RobotUnitListenerTopic", "The topic receiving events for RobotUnitListener");
            defineOptionalProperty<std::string>("DebugDrawerUpdatesTopicName", "DebugDrawerUpdates", "The topic receiving events for debug drawing");
            defineOptionalProperty<std::size_t>("PublishPeriodMs", 10, "Milliseconds between each publish");

            defineRequiredProperty<std::string>("RobotFileName", "Robot file name, e.g. robot_model.xml");
            defineOptionalProperty<std::string>("RobotFileNameProject", "", "Project in which the robot filename is located (if robot is loaded from an external project)");

            defineOptionalProperty<std::string>("KinematicUnitName", "KinematicUnit", "The name of the created kinematic unit");
            defineOptionalProperty<std::string>("RobotNodeSetName", "Robot", "Robot node set name as defined in robot xml file, e.g. 'LeftArm'");

            defineOptionalProperty<std::string>("PlatformUnitName", "PlatformUnit", "The name of the created platform unit");
            defineOptionalProperty<std::string>("PlatformName", "Platform", "Name of the platform (will publish values on PlatformName + 'State')");

            defineOptionalProperty<std::string>("ForceTorqueUnitName", "ForceTorqueUnit", "The name of the created force troque unit (empty = default)");
            defineOptionalProperty<std::string>("ForceTorqueTopicName", "ForceTorqueValues", "Name of the topic on which the force torque sensor values are provided");

            defineOptionalProperty<std::string>("InertialMeasurementUnitName", "InertialMeasurementUnit", "The name of the created inertial measurement unit (empty = default)");
            defineOptionalProperty<std::string>("IMUTopicName", "IMUValues", "Name of the IMU Topic.");

            defineOptionalProperty<bool>("DebugObserverPublishSensorValues", true, "Whether sensor values are send to the debug observer topic", PropertyDefinitionBase::eModifiable);
            defineOptionalProperty<bool>("DebugObserverPublishControlTargets", true, "Whether control targets are send to the debug observer topic", PropertyDefinitionBase::eModifiable);
            defineOptionalProperty<std::string>("DebugObserverTopicName", "DebugObserver", "The topic where updates are send to");
            defineOptionalProperty<std::uint64_t>("DebugObserverPrintEveryNthIterations", 1, "Only every nth iteration data is send to the debug observer", PropertyDefinitionBase::eModifiable);
        }
    };

    /// @brief The current state of the multi step initialization of a RobotUnit
    enum class RobotUnitState : std::size_t
    {
        PreComponentInitialization,
        InitializingDevices,
        //ok to use devices/create controllers
        InitializingUnits,
        WaitingForRTThreadInitialization,
        Running,
        Exiting
    };
}
namespace std
{
    std::string to_string(armarx::RobotUnitState s);
}

namespace armarx
{
    std::ostream& operator<<(std::ostream& o, RobotUnitState s);

    /**
     * @brief This class is a friend of LVL1Controller.
     *
     * It encapsulates all access to privates of a LVL1Controller
     * to reduce the amount of code which could break invariants
     * in LVL1Controller and to collect all such code in one place
     * to make this problem more manageable.
     *
     * Also the private functions of LVL1Controller should not be made public to prevent
     * other classes (e.g. KinematicSubUnit) or derived LVL1Controllers
     * from accessing them
     *
     */
    class RobotUnitLVL1ControllerAccess
    {
    protected:
        static void RtActivateLVL1Controller(const LVL1ControllerPtr& lvl1)
        {
            lvl1->rtActivateController();
        }
        static void RtDeactivateLVL1Controller(const LVL1ControllerPtr& lvl1)
        {
            lvl1->rtDeactivateController();
        }
        static void RtDeactivateLVL1ControllerBecauseOfError(const LVL1ControllerPtr& lvl1)
        {
            lvl1->rtDeactivateControllerBecauseOfError();
        }

        static void RtSetLVL1ControllerActive(const LVL1ControllerPtr& lvl1, bool activate, bool deactivateBecauseOfError = false)
        {
            if (activate)
            {
                RtActivateLVL1Controller(lvl1);
            }
            else
            {
                if (deactivateBecauseOfError)
                {
                    RtDeactivateLVL1ControllerBecauseOfError(lvl1);
                }
                else
                {
                    RtDeactivateLVL1Controller(lvl1);
                }
            }
        }
        static void InitLVL1Controler(
            LVL1ControllerPtr lvl1,
            RobotUnit* ru,
            std::size_t ctrlId,
            StringStringDictionary ctrlDeviceControlModeMap,
            std::vector<char> ctrlDeviceUsedBitmap,
            std::vector<std::size_t> ctrlDeviceUsedIndices,
            bool deletable
        )
        {
            lvl1->robotUnitInit(ru, ctrlId, std::move(ctrlDeviceControlModeMap), std::move(ctrlDeviceUsedBitmap), std::move(ctrlDeviceUsedIndices), deletable);
        }
        static void PublishLVL1Controller(
            LVL1ControllerPtr lvl1,
            const SensorAndControl& sac,
            const DebugDrawerInterfacePrx& draw,
            const DebugObserverInterfacePrx& observer
        )
        {
            lvl1 ->publish(sac, draw, observer);
        }
        bool GetLVL1ControllerStatusReportedActive(LVL1ControllerPtr lvl1)
        {
            return lvl1->statusReportedActive;
        }
        bool GetLVL1ControllerStatusReportedRequested(LVL1ControllerPtr lvl1)
        {
            return lvl1->statusReportedRequested;
        }
        void UpdateLVL1ControllerStatusReported(LVL1ControllerPtr lvl1)
        {
            lvl1->statusReportedActive = lvl1->isActive.load();
            lvl1->statusReportedRequested = lvl1->isRequested.load();
        }
        void SetLVL1ControllerRequested(LVL1ControllerPtr lvl1, bool requested)
        {
            lvl1->isRequested = requested;
        }

        void DeactivateLVL1ControllerPublishing(
            LVL1ControllerPtr lvl1,
            const DebugDrawerInterfacePrx& draw,
            const DebugObserverInterfacePrx& observer)
        {
            lvl1->deactivatePublish(draw, observer);
        }
    };

    class RobotUnit :
        virtual public Component,
        virtual public RobotUnitInterface,
        virtual public LVL1ControllerDescriptionProviderInterface,
        virtual public RobotUnitLVL1ControllerAccess
    {
    protected:
        using PublisherTaskT = SimplePeriodicTask<std::function<void(void)>>;
    public:
        using MutexType = std::recursive_mutex;
        using GuardType = std::unique_lock<MutexType>;
        using ClockT = std::chrono::high_resolution_clock;
        using TimePointT = std::chrono::high_resolution_clock::time_point;
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        // data
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        //devices
    private:
        /**
         * @brief ControlDevices added to this unit
         * data may only be added during and only be used after State::InitializingDevices
         */
        KeyValueVector<std::string, ControlDevicePtr> controlDevices;
        /**
         * @brief SensorDevices added to this unit
         * data may only be added during and only be used after State::InitializingDevices
         */
        KeyValueVector<std::string, SensorDevicePtr > sensorDevices;

        /// @brief const pointer to all ControlDevices (passed to GenerateConfigDescription of a LVL1Controller)
        std::map<std::string, ConstControlDevicePtr> controlDevicesConstPtr;
        /// @brief const pointer to all SensorDevices (passed to GenerateConfigDescription of a LVL1Controller)
        std::map<std::string, ConstSensorDevicePtr> sensorDevicesConstPtr;

        /// @brief a pointer to the RTThreadTimingsSensorDevice used to meassure timings in the rt loop
        RTThreadTimingsSensorDevicePtr rtThreadTimingsSensorDevice;
        /// @brief The device name of rtThreadTimingsSensorDevice
        static const std::string rtThreadTimingsSensorDeviceName;
        //triple buffers
    private:
        /// @brief Controllers the RT thread is supposed to activate (direction nonRT -> RT) (some lvl1 controllers may be null)
        /// data may only be used after State::InitializingDevices
        /// all LVL1Controller have to be sorted by id (null controllers are at the end)
        WriteBufferedTripleBuffer<LVL0AndLVL1Controllers> controllersRequested;
        /// @brief Controllers the RT thread is currently running (direction RT -> nonRT) (some lvl1 controllers may be null)
        /// data may only be used after State::InitializingDevices
        /// all LVL1Controller are sorted by id (null controllers may be anywhere)
        WriteBufferedTripleBuffer<LVL0AndLVL1Controllers> controllersActivated;
        /// @brief Transfers the current sensor values and control targets (direction RT -> nonRT)
        /// data may only be used after State::InitializingDevices
        TripleBuffer<SensorAndControl> sensorAndControl;

        //(instantiating) lvl1 controllers
    private:
        /// data may only be used after State::InitializingDevices
        std::thread::id initLVL1ThreadId;
        /// data may only be used after State::InitializingDevices
        std::map<std::string, std::string> initLVL1UsedControlTargets;
        /// @brief Holds all currently loaded LVL1 controllers (index: [instancename])(May not be accessed in rt.)
        std::map<std::string, LVL1ControllerPtr> lvl1Controllers;
        std::map<std::string, LVL1ControllerPtr> lvl1ControllersToBeDeleted;
        /// @brief the id of the next LVL1Controller instance that is created
        std::size_t lvl1ControllerNextId {0};

        //listeners
    private:
        std::string robotUnitListenerTopicName;
        RobotUnitListenerPrx listenerPrx;
        DebugDrawerInterfacePrx debugDrawerPrx;

        //publishing
    private:
        std::uint64_t publishIterationCount {0};

        IceUtil::Time publishNewSensorDataTime;
        PublisherTaskT::pointer_type publisherTask;
        std::set<std::string> lastReportedClasses;

        std::atomic_bool debugObserverPublishSensorValues;
        std::atomic_bool debugObserverPublishControlTargets;
        std::atomic<std::uint64_t> debugObserverSkipIterations;

        std::chrono::microseconds lastDebugObserverFlush;
        std::chrono::microseconds lastPublishLoop;
        DebugObserverInterfacePrx debugObserverPrx;


        //other
    private:
        /// @brief The RobotUnit's state
        std::atomic<RobotUnitState> state {RobotUnitState::PreComponentInitialization};
        std::map<std::string, DynamicLibraryPtr> loadedLibs;
        std::vector<ManagedIceObjectPtr> units;
        std::thread::id rtThreadId;
        mutable MutexType dataMutex;
        static const std::set<RobotUnitState> devicesReadyStates;

        //robot
    private:
        boost::shared_ptr<VirtualRobot::Robot> robot;
        std::string robotNodeSetName;
        std::string robotProjectName;
        std::string robotFileName;
        std::string robotPlatformName;

        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        // util
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
    protected:
        template<class ExceptT = LogicError>
        inline void throwIfStateIsNot(const std::set<RobotUnitState>& sinit, const std::string& fnc) const;
        template<class ExceptT = LogicError>
        inline void throwIfStateIsNot(RobotUnitState s, const std::string& fnc) const;
        template<class ExceptT = LogicError>
        inline void throwIfDevicesNotReady(const std::string& fnc) const;

        template<class Cont>
        static Ice::StringSeq GetNonNullNames(const Cont& c);

        template<class PtrT>
        void checkRefCountIsOne(const PtrT& p, const std::string& name) const;
        GuardType getGuard() const;
        inline const KeyValueVector<std::string, SensorDevicePtr>& getSensorDevices() const;
        inline const KeyValueVector<std::string, ControlDevicePtr>& getControlDevices() const;

        void checkLVL1ControllerClassName(const std::string& className) const;

        //time
        static TimePointT Now()
        {
            return ClockT::now();
        }
        static std::chrono::microseconds MicroToNow(TimePointT last)
        {
            return std::chrono::duration_cast<std::chrono::microseconds>(Now() - last);
        }
        static std::chrono::microseconds MicroNow();

        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        // other
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
    public:
        /// @see PropertyUser::createPropertyDefinitions()
        virtual armarx::PropertyDefinitionsPtr createPropertyDefinitions() override;
        virtual std::string getDefaultName() const override
        {
            return "RobotUnit";
        }
        // PropertyUser interface
    public:
        virtual void icePropertiesUpdated(const std::set<std::string>& changedProperties) override;
    protected:
        virtual void onInitRobotUnit()       {}
        virtual void onConnectRobotUnit()    {}
        virtual void onDisconnectRobotUnit() {}
        /// @brief called in RobotUnit::onExitComponent() before calling RobotUnit::finishRunning
        virtual void onExitRobotUnit()       {}
        virtual void onInitComponent()       final;
        virtual void onConnectComponent()    final;
        virtual void onDisconnectComponent() final;
        virtual void onExitComponent()       final;

        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        // TripleBuffer
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
    protected:
        //rt requested
        bool rtRequestedControllersChanged() const;
        const LVL0AndLVL1Controllers&       rtGetRequestedControllers()    const;
        const std::vector<LVL0Controller*>& rtGetRequestedLVL0Controllers() const;
        const std::vector<LVL1Controller*>& rtGetRequestedLVL1Controllers() const;
        // rt activated
        void rtWriteActivatedControllers();
        LVL0AndLVL1Controllers&       rtGetActivatedControllers();
        std::vector<LVL0Controller*>& rtGetActivatedLVL0Controllers();
        std::vector<LVL1Controller*>& rtGetActivatedLVL1Controllers();
        // rt sensor and control
        //declaration further down
        //void rtUpdateSensorAndControlBuffer(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration);

        // non rt requested
        void writeRequestedControllers(LVL0AndLVL1Controllers&& setOfControllers);
        const LVL0AndLVL1Controllers&       getRequestedControllers()     const;
        const std::vector<LVL0Controller*>& getRequestedLVL0Controllers() const;
        const std::vector<LVL1Controller*>& getRequestedLVL1Controllers() const;
        // non rt activated
        bool activatedControllersChanged() const;
        const LVL0AndLVL1Controllers&       getActivatedControllers()     const;
        const std::vector<LVL0Controller*>& getActivatedLVL0Controllers() const;
        const std::vector<LVL1Controller*>& getActivatedLVL1Controllers() const;
        // non rt sensor and control
        bool sensorAndControlBufferChanged() const;
        const SensorAndControl& getSensorAndControlBuffer() const;

        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        // State: InitializingDevices
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
    protected:
        void addControlDevice(const ControlDevicePtr& cd);
        void addSensorDevice(const SensorDevicePtr& sd);

        virtual RTThreadTimingsSensorDevicePtr createRTThreadTimingSensorDevice() const;

        void finishDeviceInitialization();

        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        // State: InitializingUnits
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
    protected:
        virtual void initializeDefaultUnits();
        virtual void initializeKinematicUnit();
        virtual void initializePlatformUnit();
        virtual void initializeForceTorqueUnit();
        virtual void initializeInertialMeasurementUnit();
        void addUnit(ManagedIceObjectPtr unit);
        void finishUnitInitialization();
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        // State: Running (RT)
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        virtual void initRTThread();

        //switching helper (rt)
        //is in the base class RobotUnitLVL1ControllerAccess that is a friend of LVL1Controller
        //static void RtSetLVL1ControllerActive(const LVL1ControllerPtr& lvl1, bool activate = 1, bool deactivateBecauseOfError = false);
        void rtDeactivateAssignedLVL1Controller(std::size_t ctrlDevIndex, bool writeActivatedControlers = true);

        //run (these functions should be called in the control loop)
        bool rtSwitchControllerSetup();
        void rtRunLVL1Controllers(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration);
        void rtHandleInvalidTargets();
        /**
         * @brief Runs LVL0 controllers and writes target values to ControlDevices
         * @param sensorValuesTimestamp
         * @param timeSinceLastIteration
         */
        void rtRunLVL0Controllers(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration);
        void rtReadSensorDeviceValues(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration);
        //rtBusSendReceive call this robot specific function here
        void rtUpdateSensorAndControlBuffer(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration);
        void rtResetAllTargets();

        RTThreadTimingsSensorDevice& rtGetRTThreadTimingsSensorDevice()
        {
            return *rtThreadTimingsSensorDevice;
        }

        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        // State: InitializingUnits + Running (NON RT)
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
    public:
        //create lvl1 controllers
        virtual ConstSensorDevicePtr   getSensorDevice(const std::string& sensorDeviceName) const;
        virtual ConstControlDevicePtr  getControlDevice(const std::string& deviceName);
        virtual ControlTargetBase* getControlTarget(const std::string& deviceName, const std::string& controlMode) override;
        virtual LVL1ControllerInterfacePrx createLVL1Controller(const std::string& className, const std::string& instanceName, const LVL1ControllerConfigPtr& config, const Ice::Current&) override;
        virtual LVL1ControllerInterfacePrx createLVL1ControllerFromVariantConfig(const std::string& className, const std::string& instanceName, const StringVariantBaseMap& variants, const Ice::Current&) override;
        LVL1ControllerPtr createLVL1Controller(const std::string& className, const std::string& instanceName, const LVL1ControllerConfigPtr& config, bool deletable);
        //deleting lvl1 controllers
        virtual void deleteLVL1Controller(const std::string& name, const Ice::Current& = GlobalIceCurrent) override;
        virtual void deleteLVL1Controllers(const Ice::StringSeq& names, const Ice::Current& = GlobalIceCurrent) override;
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        // State: Running (NON RT)
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
    public:
        //switching
        virtual void activateLVL1Controller(const std::string& name, const Ice::Current& = GlobalIceCurrent) override;
        virtual void activateLVL1Controllers(const Ice::StringSeq& names, const Ice::Current& = GlobalIceCurrent) override;
        virtual void deactivateLVL1Controller(const std::string& name, const Ice::Current& = GlobalIceCurrent) override;
        virtual void deactivateLVL1Controllers(const Ice::StringSeq& names, const Ice::Current& = GlobalIceCurrent) override;
        virtual void switchLVL1ControllerSetup(const Ice::StringSeq& newSetup, const Ice::Current& = GlobalIceCurrent) override;
    protected:
        void setActivateControllersRequest(std::set<LVL1ControllerPtr, LVL1Controller::IdComp>&& ctrls);
        /**
         * @brief This hook is called prior to writing the set of requested controllers.
         * A deriving class can override this function to perform some additional checks
         * (e.g. two controll devices may always need to have the same control mode, this can be checked here)
         * If any check fails, throw an exception with a descriptive message.
         * @param controllers The controllers to activate
         */
        virtual void checkSetOfControllersToActivate(LVL0AndLVL1Controllers /*controllers*/) const {}
        // publishing
    protected:
        virtual void publish();

        //shutting down
    protected:
        ///@brief joint your thread in this hook. (used by RobotUnit::finishRunning())
        virtual void joinRTThread() = 0;
        ///@brief delete all units (used by RobotUnit::finishRunning())
        virtual void removeAllUnits();
    private:
        void finishRunning();

        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        // Lib loading
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
    public:
        virtual bool loadLibFromPath(const std::string& path, const Ice::Current& = GlobalIceCurrent) override;
        virtual bool loadLibFromPackage(const std::string& package, const std::string& lib, const Ice::Current& = GlobalIceCurrent) override;
    private:
        bool loadLibFromAbsolutePath(const std::string& path);
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        // querry data
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
    public:
        //lvl1
        virtual Ice::StringSeq getLVL1ControllerNames(const Ice::Current& = GlobalIceCurrent) const override;
        virtual Ice::StringSeq getRequestedLVL1ControllerNames(const Ice::Current& = GlobalIceCurrent) const override;
        virtual Ice::StringSeq getActivatedLVL1ControllerNames(const Ice::Current& = GlobalIceCurrent) const override;

        virtual LVL1ControllerInterfacePrx getLVL1Controller(const std::string& name, const Ice::Current&) const override;
        virtual StringLVL1ControllerPrxDictionary getAllLVL1Controllers(const Ice::Current&) const override;
        std::vector<LVL1ControllerPtr> getLVL1ControllersNotNull(const std::vector<std::string>& names) const;
        LVL1ControllerPtr getLVL1ControllerNotNull(const std::string& name) const;

        virtual LVL1ControllerStatus getLVL1ControllerStatus(const std::string& name, const Ice::Current& = GlobalIceCurrent) const override;
        virtual LVL1ControllerStatusSeq getLVL1ControllerStatuses(const Ice::Current&) const override;

        virtual LVL1ControllerDescription getLVL1ControllerDescription(const std::string& name, const Ice::Current& = GlobalIceCurrent) const override;
        virtual LVL1ControllerDescriptionSeq getLVL1ControllerDescriptions(const Ice::Current&) const override;

        virtual LVL1ControllerDescriptionWithStatus getLVL1ControllerDescriptionWithStatus(const std::string& name, const Ice::Current& = GlobalIceCurrent) const override;
        virtual LVL1ControllerDescriptionWithStatusSeq getLVL1ControllerDescriptionsWithStatuses(const Ice::Current&) const override;
        //control devices
        virtual Ice::StringSeq getControlDeviceNames(const Ice::Current&) const override;
        ControlDeviceDescription getControlDeviceDescription(std::size_t cdidx) const;
        virtual ControlDeviceDescription getControlDeviceDescription(const std::string& name, const Ice::Current& = GlobalIceCurrent) const override;
        virtual ControlDeviceDescriptionSeq getControlDeviceDescriptions(const Ice::Current&) const override;
        ControlDeviceStatus getControlDeviceStatus(std::size_t cdidx) const;
        virtual ControlDeviceStatus getControlDeviceStatus(const std::string& name, const Ice::Current& = GlobalIceCurrent) const override;
        virtual ControlDeviceStatusSeq getControlDeviceStatuses(const Ice::Current&) const override;

        //sensor devices
        virtual Ice::StringSeq getSensorDeviceNames(const Ice::Current&) const override;
        SensorDeviceDescription getSensorDeviceDescription(std::size_t cdidx) const;
        virtual SensorDeviceDescription getSensorDeviceDescription(const std::string& name, const Ice::Current& = GlobalIceCurrent) const override;
        virtual SensorDeviceDescriptionSeq getSensorDeviceDescriptions(const Ice::Current&) const override;
        SensorDeviceStatus getSensorDeviceStatus(std::size_t idx) const;
        virtual SensorDeviceStatus getSensorDeviceStatus(const std::string& name, const Ice::Current& = GlobalIceCurrent) const override;
        virtual SensorDeviceStatusSeq getSensorDeviceStatuses(const Ice::Current&) const override;

        //lvl1controller classes
        virtual Ice::StringSeq getLVL1ControllerClassNames(const Ice::Current& = GlobalIceCurrent) const override;
        virtual LVL1ControllerClassDescription getLVL1ControllerClassDescription(const std::string& className, const Ice::Current& = GlobalIceCurrent) const override;
        virtual LVL1ControllerClassDescriptionSeq getLVL1ControllerClassDescriptions(const Ice::Current&) const override;

        //units
        virtual  Ice::ObjectProxySeq getUnits(const Ice::Current& = GlobalIceCurrent) const override;
        virtual Ice::ObjectPrx getUnit(const std::string& staticIceId, const Ice::Current&) const override;
        const ManagedIceObjectPtr& getUnit(const std::string& staticIceId) const;
        virtual           KinematicUnitInterfacePrx getKinematicUnit(const Ice::Current& = GlobalIceCurrent) const override;
        virtual         ForceTorqueUnitInterfacePrx getForceTorqueUnit(const Ice::Current& = GlobalIceCurrent) const override;
        virtual InertialMeasurementUnitInterfacePrx getInertialMeasurementUnit(const Ice::Current& = GlobalIceCurrent) const override;
        virtual            PlatformUnitInterfacePrx getPlatformUnit(const Ice::Current& = GlobalIceCurrent) const override;
        //other ice functions
        virtual std::string getListenerTopicName(const Ice::Current& = GlobalIceCurrent) const override;
        //other local functions
        const boost::shared_ptr<VirtualRobot::Robot>& getRobot() const;
        std::string getRobotName() const;
        const std::string& getRobotNodetSeName() const
        {
            return robotNodeSetName;
        }
        const std::string& getRobotProjectName() const
        {
            return robotProjectName;
        }
        const std::string& getRobotFileName() const
        {
            return robotFileName;
        }
        const std::string& getRobotPlatformName() const
        {
            return robotPlatformName;
        }

        std::size_t getNumberOfControlDevices() const;
        std::size_t getNumberOfSensorDevices() const;
        std::size_t rtGetControllerBufferSize() const
        {
            return controlDevices.size();
        }
        std::size_t rtGetNumberOfSensorDevices() const
        {
            return sensorDevices.size();
        }
        std::size_t getControllerBufferSize() const
        {
            return getNumberOfControlDevices();
        }

        const DebugDrawerInterfacePrx& getDebugDrawerProxy() const;


        bool areDevicesReady() const
        {
            return devicesReadyStates.count(state);
        }

        RobotUnitState getRobotUnitState() const
        {
            return state;
        }
        virtual std::string getName() const override
        {
            return Component::getName();
        }
    protected:
        const RobotUnitListenerPrx& getListenerProxy() const;
    };

    const KeyValueVector<std::string, SensorDevicePtr>& RobotUnit::getSensorDevices() const
    {
        return sensorDevices;
    }

    const KeyValueVector<std::string, ControlDevicePtr>& RobotUnit::getControlDevices() const
    {
        return controlDevices;
    }

    template<class ExceptT>
    void RobotUnit::throwIfStateIsNot(const std::set<RobotUnitState>& sinit, const std::string& fnc) const
    {
        std::set<RobotUnitState> sset {sinit};
        if (! sset.count(state))
        {
            std::stringstream ss;
            ss << fnc << ": Can't be called if state is not in (";
            bool fst = true;
            for (RobotUnitState st : sset)
            {
                if (!fst)
                {
                    ss << ", ";
                }
                ss << st;
                fst = false;
            }
            ss << " (current state " << getRobotUnitState() << ")";
            ARMARX_ERROR << ss.str();
            throw ExceptT {ss.str()};
        }
    }
    template<class ExceptT>
    void RobotUnit::throwIfStateIsNot(RobotUnitState s, const std::string& fnc) const
    {
        throwIfStateIsNot<ExceptT>(std::set<RobotUnitState> {s}, fnc);
    }
    template<class ExceptT>
    void RobotUnit::throwIfDevicesNotReady(const std::string& fnc) const
    {
        throwIfStateIsNot<ExceptT>(devicesReadyStates, fnc);
    }

    template<class Cont>
    Ice::StringSeq RobotUnit::GetNonNullNames(const Cont& c)
    {
        Ice::StringSeq result;
        result.reserve(c.size());
        for (const auto& e : c)
        {
            if (e)
            {
                result.emplace_back(e->getName());
            }
        }
        return result;
    }

    template<class PtrT>
    void RobotUnit::checkRefCountIsOne(const PtrT& p, const std::string& name) const
    {
        const auto uc = p->__getRef();
        if (uc != 1)
        {
            ARMARX_WARNING << "Pointer " << name << " should have usecount 1 but instead hase usecount " << uc;
        }
    }
}
#endif
