/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI
 * @author     Raphael ( raphael dot grimm at student dot kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */
#ifndef _ARMARX_LIB_RobotAPI_SensorValueBase_H
#define _ARMARX_LIB_RobotAPI_SensorValueBase_H

#include <typeinfo>
#include <memory>
#include <string>
#include <map>

#include <ArmarXCore/observers/variant/TimedVariant.h>

#include "../util.h"

namespace armarx
{
    class SensorValueBase
    {
    public:
        virtual void copyTo(SensorValueBase* target) const = 0;
        virtual std::unique_ptr<SensorValueBase> clone() const = 0;
        virtual std::string getSensorValueType(bool withoutNamespaceSpecifier) const = 0;

        /// @brief used to send the data to the DebugObserverTopic and to other Components (e.g. GUI widgets)
        virtual std::map<std::string, VariantBasePtr>   toVariants(const IceUtil::Time& timestamp) const
        {
            return {};
        }

        template<class T>
        bool isA() const
        {
            return dynamic_cast<const T*>(this);
        }

        template<class T>
        const T* asA() const
        {
            return dynamic_cast<const T*>(this);
        }

        template<class T>
        T* asA()
        {
            return dynamic_cast<T*>(this);
        }

        template<class T, class = typename std::enable_if<std::is_base_of<SensorValueBase, T>::value>::type>
        void copyTo(std::unique_ptr<T>& target) const
        {
            copyTo(target.get());
        }

        virtual ~SensorValueBase() = default;
    };
}

#define DETAIL_SensorValueBase_DEFAULT_METHOD_IMPLEMENTATION                                        \
    using SensorValueBase = ::armarx::SensorValueBase;                                              \
    virtual void copyTo(SensorValueBase* target) const override                                     \
    {                                                                                               \
        using this_type = typename std::decay<decltype(*this)>::type;                               \
        *(target->asA<this_type>()) = *this;                                                        \
    }                                                                                               \
    virtual std::unique_ptr<SensorValueBase> clone() const override                                 \
    {                                                                                               \
        using this_type = typename std::decay<decltype(*this)>::type;                               \
        std::unique_ptr<SensorValueBase> c {new this_type};                                         \
        SensorValueBase::copyTo(c);                                                                 \
        return c;                                                                                   \
    }                                                                                               \
    virtual std::string getSensorValueType(bool withoutNamespaceSpecifier = false) const override   \
    {                                                                                               \
        return getTypeName(*this, withoutNamespaceSpecifier);                                       \
    }

namespace armarx
{
    class SensorValueDummy : public SensorValueBase
    {
    public:
        DETAIL_SensorValueBase_DEFAULT_METHOD_IMPLEMENTATION
    };
}

#endif
