/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::KinematicSubUnit
 * @author     Raphael ( raphael dot grimm at student dot kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#ifndef _ARMARX_LIB_RobotAPI_KinematicSubUnit_H
#define _ARMARX_LIB_RobotAPI_KinematicSubUnit_H

#include <mutex>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <RobotAPI/components/units/KinematicUnit.h>

#include "RobotUnitSubUnit.h"
#include "../util.h"
#include "../SensorValues/SensorValue1DoFActuator.h"
#include "../LVL1Controllers/LVL1KinematicUnitPassThroughController.h"

namespace armarx
{
    class RobotUnit;

    TYPEDEF_PTRS_HANDLE(KinematicSubUnit);
    class KinematicSubUnit:
        virtual public RobotUnitSubUnit,
        virtual public KinematicUnit
    {
    public:
        struct ActuatorData
        {
            std::string name;
            std::size_t sensorDeviceIndex;
            LVL1KinematicUnitPassThroughControllerPtr ctrlPos;
            LVL1KinematicUnitPassThroughControllerPtr ctrlVel;
            LVL1KinematicUnitPassThroughControllerPtr ctrlTor;

            const LVL1ControllerPtr getController(ControlMode c) const;
        };

        void setupData(std::string relRobFile, VirtualRobot::RobotPtr rob, std::map<std::string, ActuatorData> newDevs, RobotUnit* newRobotUnit);
        virtual void update(const SensorAndControl& sc, const LVL0AndLVL1Controllers& c) override;

        // KinematicUnitInterface interface
        virtual void requestJoints(const Ice::StringSeq&, const Ice::Current&) override;
        virtual void releaseJoints(const Ice::StringSeq&, const Ice::Current&) override;

        virtual void setJointAngles(const NameValueMap& angles, const Ice::Current&) override;
        virtual void setJointVelocities(const NameValueMap& velocities, const Ice::Current&) override;
        virtual void setJointTorques(const NameValueMap& torques, const Ice::Current&) override;
        virtual void switchControlMode(const NameControlModeMap& ncm, const Ice::Current&) override;

        virtual void setJointAccelerations(const NameValueMap&, const Ice::Current&) override;
        virtual void setJointDecelerations(const NameValueMap&, const Ice::Current&) override;

        virtual NameControlModeMap getControlModes(const Ice::Current&) override;

        // KinematicUnit interface
        virtual void onInitKinematicUnit()  override {}
        virtual void onStartKinematicUnit() override {}
        virtual void onExitKinematicUnit()  override {}
    private:
        std::map<std::string, ActuatorData> devs;
        // never use this when holding the mutex! (could cause deadlocks)
        RobotUnit* robotUnit;
        mutable std::mutex dataMutex;
        NameControlModeMap ctrlModes;
        NameValueMap ang;
        NameValueMap vel;
        NameValueMap acc;
        NameValueMap tor;
        NameValueMap motorCurrents;
        NameValueMap motorTemperatures;

        template<class SensorValueT, float SensorValueT::* Member>
        static void UpdateNameValueMap(NameValueMap& nvm, const std::string& name, const SensorValueBase* sv, bool& changeState)
        {
            const SensorValueT* s = sv->asA<SensorValueT>();
            if (!s)
            {
                return;
            };
            const float v = s->*Member;
            if (!nvm.count(name) || (nvm.at(name) != v))
            {
                changeState = true;
            }
            nvm[name] = v;
        }
    };
}
#endif
