/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::PlatformSubUnit
 * @author     Raphael ( raphael dot grimm at student dot kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */
#include "PlatformSubUnit.h"

void armarx::PlatformSubUnit::update(const armarx::SensorAndControl& sc, const LVL0AndLVL1Controllers&)
{
    if(!getProxy())
    {
        //this unit is not initialized yet
        ARMARX_IMPORTANT << deactivateSpam(1) << "not initialized yet";
        return;
    }
    if(!listenerPrx)
    {
        ARMARX_IMPORTANT << deactivateSpam(1) << "listener is not set";
        return;
    }
    const SensorValueBase* sensorValue = sc.sensors.at(platformSensorIndex).get();
    ARMARX_CHECK_EXPRESSION(sensorValue);
    std::lock_guard<std::mutex> guard {dataMutex};
    //pos
    {
        ARMARX_CHECK_EXPRESSION(sensorValue->isA<SensorValueHolonomicPlatformRelativePosition>());
        const SensorValueHolonomicPlatformRelativePosition* s = sensorValue->asA<SensorValueHolonomicPlatformRelativePosition>();
        relX = s->relativePositionX;
        relY = s->relativePositionY;
        relR = s->relativePositionRotation;

        const Eigen::Matrix4f glob = getCorrectedPose();
        const float globR = VirtualRobot::MathTools::eigen4f2rpy(glob)(2);
        listenerPrx->reportPlatformPose(glob(0, 3), glob(1, 3), globR);
    }
    //vel
    {
        ARMARX_CHECK_EXPRESSION(sensorValue->isA<SensorValueHolonomicPlatformVelocity>());
        const SensorValueHolonomicPlatformVelocity* s = sensorValue->asA<SensorValueHolonomicPlatformVelocity>();
        listenerPrx->reportPlatformVelocity(s->velocityX, s->velocityY , s->velocityRotation);
    }
}

void armarx::PlatformSubUnit::move(Ice::Float vx, Ice::Float vy, Ice::Float vr, const Ice::Current&)
{
    //holding the mutex here could deadlock
    if (!pt->isControllerActive())
    {
        pt->activateController();
    }
    std::lock_guard<std::mutex> guard {dataMutex};
    pt->setVelocites(
        boost::algorithm::clamp(vx, -maxVLin, maxVLin),
        boost::algorithm::clamp(vy, -maxVLin, maxVLin),
        boost::algorithm::clamp(vr, -maxVAng, maxVAng)
    );
}

void armarx::PlatformSubUnit::moveTo(Ice::Float rx, Ice::Float ry, Ice::Float rr, Ice::Float lac, Ice::Float rac, const Ice::Current &)
{
    ARMARX_WARNING << "NYI";
}

void armarx::PlatformSubUnit::moveRelative(Ice::Float rx, Ice::Float ry, Ice::Float rr, Ice::Float lac, Ice::Float rac, const Ice::Current &)
{
    ARMARX_WARNING << "NYI";
}

void armarx::PlatformSubUnit::setMaxVelocities(Ice::Float mxVLin, Ice::Float mxVAng, const Ice::Current&)
{
    std::lock_guard<std::mutex> guard {dataMutex};
    maxVLin = std::abs(mxVLin);
    maxVAng = std::abs(mxVAng);
}

void armarx::PlatformSubUnit::setGlobalPose(armarx::PoseBasePtr globalPose, const Ice::Current&)
{
    std::lock_guard<std::mutex> guard {dataMutex};
    PosePtr p = PosePtr::dynamicCast(globalPose);
    relPositionCorrection = p->toEigen() * getRelativePose().inverse();
}

armarx::PoseBasePtr armarx::PlatformSubUnit::getGlobalPose(const Ice::Current&)
{
    std::lock_guard<std::mutex> guard {dataMutex};
    return new Pose {getCorrectedPose()};
}

void armarx::PlatformSubUnit::stopPlatform(const Ice::Current& c)
{
    move(0, 0, 0);
}

Eigen::Matrix4f armarx::PlatformSubUnit::getRelativePose() const
{
    Eigen::Matrix4f rp = VirtualRobot::MathTools::rpy2eigen4f(0, 0, relR);
    rp(0, 3) = relX;
    rp(1, 3) = relY;
    return rp;
}
