/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * \package    RobotAPI::gui-plugins::RobotUnitPlugin
 * \author     Raphael ( ufdrv at student dot kit dot edu )
 * \date       2017
 * \copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */
#include "RobotUnitLVL1ControllerClassesWidget.h"
#include "ui_RobotUnitLVL1ControllerClassesWidget.h"

#include <QGridLayout>

namespace armarx
{
    RobotUnitLVL1ControllerClassesWidget::RobotUnitLVL1ControllerClassesWidget(QWidget* parent) :
        QWidget(parent),
        ui(new Ui::RobotUnitLVL1ControllerClassesWidget)
    {
        ui->setupUi(this);
    }

    RobotUnitLVL1ControllerClassesWidget::~RobotUnitLVL1ControllerClassesWidget()
    {
        delete ui;
    }

    void RobotUnitLVL1ControllerClassesWidget::lvl1ControllerClassAdded(std::string name)
    {
        std::lock_guard<std::recursive_mutex> guard {mutex};
        add(robotUnit->getLVL1ControllerClassDescription(name));
    }

    void RobotUnitLVL1ControllerClassesWidget::reset(RobotUnitInterfacePrx ru)
    {
        std::lock_guard<std::recursive_mutex> guard {mutex};
        robotUnit = ru;
        //delete all
        ui->treeWidget->clear();
        entries.clear();
        //add filter line
        {
            QTreeWidgetItem* filterHeader = new QTreeWidgetItem;
            ui->treeWidget->addTopLevelItem(filterHeader);
            filterHeader->setText(0, "Filter");

            QTreeWidgetItem* filter = new QTreeWidgetItem;
            filterHeader->addChild(filter);
            filter->setFirstColumnSpanned(true);
            QWidget* w = new QWidget;
            QGridLayout* l = new QGridLayout;
            w->setLayout(l);
            l->setContentsMargins(0, 0, 0, 0);
            l->addItem(new QSpacerItem {0, 0, QSizePolicy::MinimumExpanding}, 0, 3);
            l->addWidget(new QLabel {"Combine filters with"}, 0, 0, 1, 1);
            filterCombination = new QComboBox;
            filterCombination->addItem("Or");
            filterCombination->addItem("And");
            l->addWidget(filterCombination, 0, 1, 1, 1);

            filterNameActive = new QCheckBox;
            filterNameActive->setText("Filter by name");
            l->addWidget(filterNameActive, 1, 0, 1, 1);

            filterName = new QLineEdit;
            l->addWidget(filterName, 1, 1, 1, 1);

            filterNameInverted = new QCheckBox;
            filterNameInverted->setText("Invert filter");
            l->addWidget(filterNameInverted, 1, 2, 1, 1);
            connect(filterNameActive, SIGNAL(toggled(bool)), filterName, SLOT(setEnabled(bool)));
            connect(filterNameActive, SIGNAL(toggled(bool)), filterNameInverted, SLOT(setEnabled(bool)));
            filterName->setEnabled(false);
            filterNameInverted->setEnabled(false);

            filterRemoteCreationActive = new QCheckBox;
            filterRemoteCreationActive->setText("Filter by remote creation capabilities");
            l->addWidget(filterRemoteCreationActive, 2, 0, 1, 1);
            filterRemoteCreation = new QComboBox;
            filterRemoteCreation->addItem("Both");
            filterRemoteCreation->addItem("With");
            filterRemoteCreation->addItem("Without");
            l->addWidget(filterRemoteCreation, 2, 1, 1, 1);
            connect(filterRemoteCreationActive, SIGNAL(toggled(bool)), filterRemoteCreation, SLOT(setEnabled(bool)));
            filterRemoteCreationActive->setEnabled(false);

            connect(filterCombination, SIGNAL(currentIndexChanged(QString)), this, SLOT(filterUpdated()));
            connect(filterNameActive, SIGNAL(clicked(bool)), this, SLOT(filterUpdated()));
            connect(filterRemoteCreationActive, SIGNAL(clicked(bool)), this, SLOT(filterUpdated()));
            connect(filterNameInverted, SIGNAL(clicked(bool)), this, SLOT(filterUpdated()));
            connect(filterName, SIGNAL(textChanged(QString)), this, SLOT(filterUpdated()));
            connect(filterRemoteCreation, SIGNAL(currentIndexChanged(int)), this, SLOT(filterUpdated()));

            filterNameActive->setChecked(false);
            filterRemoteCreationActive->setChecked(false);

            ui->treeWidget->setItemWidget(filter, 0, w);
        }
        //fill
        if (robotUnit)
        {
            for (LVL1ControllerClassDescription& desc : robotUnit->getLVL1ControllerClassDescriptions())
            {
                add(desc);
            }
        }
    }

    void RobotUnitLVL1ControllerClassesWidget::filterUpdated()
    {
        for (auto& entry : entries)
        {
            RobotUnitLVL1ControllerClassesWidgetEntry* item = entry.second;
            if (! filterNameActive->isChecked() && !filterRemoteCreationActive->isChecked())
            {
                item->setVisible(true);
                return;
            }
            const bool combineOr = (filterCombination->currentText() == "Or");
            bool showName = !combineOr; //init to neutral element
            bool showRemote = !combineOr;//init to neutral element

            if (filterNameActive->isChecked())
            {
                showName = (item->matchName(filterName->text()) != filterNameInverted->isChecked());
            }
            if (filterRemoteCreationActive->isChecked())
            {
                if (item->hasRemoteCreation())
                {
                    showRemote = (filterRemoteCreation->currentText() != "Without");
                }
                else
                {
                    showRemote = (filterRemoteCreation->currentText() != "With");
                }
            }
            if (combineOr)
            {
                item->setVisible(showName || showRemote);
            }
            else
            {
                item->setVisible(showName && showRemote);
            }
        }

    }

    void RobotUnitLVL1ControllerClassesWidget::add(const  LVL1ControllerClassDescription& desc)
    {
        std::lock_guard<std::recursive_mutex> guard {mutex};
        if (entries.count(desc.className))
        {
            return;
        }
        entries[desc.className] = new RobotUnitLVL1ControllerClassesWidgetEntry(*this, *(ui->treeWidget), desc, robotUnit);
    }

    RobotUnitLVL1ControllerClassesWidgetEntry::RobotUnitLVL1ControllerClassesWidgetEntry(
        RobotUnitLVL1ControllerClassesWidget& parent,
        QTreeWidget& treeWidget,
        const LVL1ControllerClassDescription& desc,
        RobotUnitInterfacePrx robotUnit
    ) :
        QObject {&parent},
            className {desc.className},
            classNameQStr {QString::fromStdString(className)},
    robotUnit {robotUnit}
    {
        header = new QTreeWidgetItem {{QString::fromStdString(desc.className)}};
        treeWidget.addTopLevelItem(header);

        if (desc.configDescription)
        {
            QTreeWidgetItem* child = new QTreeWidgetItem;
            header->addChild(child);
            //enclosing widget
            QWidget* w = new QWidget;
            QVBoxLayout* wl = new QVBoxLayout;
            w->setLayout(wl);
            wl->setContentsMargins(0, 0, 0, 0);
            //header
            QHBoxLayout* l = new QHBoxLayout;
            l->setContentsMargins(0, 0, 0, 0);
            nameEdit = new QLineEdit;
            l->addWidget(new QLabel {"Controller name"});
            l->addWidget(nameEdit);
            l->addItem(new QSpacerItem {0, 0, QSizePolicy::MinimumExpanding});
            QPushButton* button = new QPushButton {"Create"};
            l->addWidget(button);
            connect(button, SIGNAL(clicked(bool)), this, SLOT(createCtrl()));
            wl->addLayout(l);
            //base
            creator = WidgetDescription::makeDescribedWidget(desc.configDescription);
            creator->layout()->setContentsMargins(0, 0, 0, 0);
            w->layout()->addWidget(creator);
            //add
            treeWidget.setItemWidget(child, 0, w);
        }
        else
        {
            QWidget* w = new QWidget;
            w->setLayout(new QHBoxLayout);
            w->layout()->setContentsMargins(0, 0, 0, 0);
            w->layout()->addWidget(new QLabel {QString::fromStdString(desc.className)});
            w->layout()->addItem(new QSpacerItem {0, 0, QSizePolicy::MinimumExpanding});
            w->layout()->addWidget(new QLabel {"No remote creation allowed."});
            header->setText(0, "");
            treeWidget.setItemWidget(header, 0, w);
        }

    }

    bool RobotUnitLVL1ControllerClassesWidgetEntry::matchName(const QString& name)
    {
        return classNameQStr.contains(name, Qt::CaseInsensitive);
    }

    bool RobotUnitLVL1ControllerClassesWidgetEntry::hasRemoteCreation()
    {
        return creator;
    }

    void RobotUnitLVL1ControllerClassesWidgetEntry::setVisible(bool vis)
    {
        header->setHidden(!vis);
    }

    void RobotUnitLVL1ControllerClassesWidgetEntry::createCtrl()
    {
        const auto instanceName = nameEdit->text().toStdString();
        const auto variants = creator->getVariants();
        if (variants.empty())
        {
            ARMARX_INFO << "creating " << instanceName << " of class " << className << " with no parameters\n";
        }
        else
        {
            std::stringstream ss;
            ss  << "creating " << instanceName << " of class " << className << " with parameters:\n";
            for (const auto& pair : variants)
            {
                if (pair.second)
                {

                    if (pair.second->data)
                    {
                        ss << "    '" << pair.first << "' of type " << pair.second->data->ice_id() << "\n";
                    }
                    else
                    {
                        ss << "    '" << pair.first << "' nullptr data \n";
                    }
                }
                else
                {
                    ss << "    '" << pair.first << "' nullptr\n";
                }
            }
            ARMARX_INFO << ss.str();
        }
        robotUnit->createLVL1ControllerFromVariantConfig(className, instanceName, variants);
    }

}
