/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * \package    RobotAPI::gui-plugins::RobotUnitPluginWidgetController
 * \author     Raphael ( ufdrv at student dot kit dot edu )
 * \date       2017
 * \copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "RobotUnitPluginWidgetController.h"
#include "StyleSheets.h"

#include <string>

#include <boost/filesystem.hpp>

#include <QDir>

#include <ArmarXCore/core/system/cmake/CMakePackageFinder.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

using namespace armarx;

RobotUnitPluginWidgetController::RobotUnitPluginWidgetController()
{
    widget.setupUi(getWidget());

    connect(widget.pushButtonLoadLib, SIGNAL(released()), this, SLOT(loadLibClicked()));

    //add subwidget ControlDevices
    {
        controlDevices = new RobotUnitControlDevicesWidget;
        //header
        auto header = makeHeaderItem("ControlDevices");
        connect(std::get<2>(header), SIGNAL(clicked(bool)), this, SLOT(refreshControlDevicesClicked()));
        connect(std::get<1>(header), SIGNAL(clicked(bool)), controlDevices, SLOT(setShown(bool)));
        connect(std::get<1>(header), SIGNAL(clicked(bool)), this, SLOT(changeVisibleCount()));
        widget.verticalLayout->addWidget(std::get<0>(header));
        widget.verticalLayout->addWidget(controlDevices);
        controlDevices->setVisible(false);
        controlDevices->setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::MinimumExpanding);
    }
    //add subwidget SensorDevices
    {
        sensorDevices = new RobotUnitSensorDevicesWidget;
        //header
        auto header = makeHeaderItem("SensorDevices");
        connect(std::get<2>(header), SIGNAL(clicked(bool)), this, SLOT(refreshSensorDevicesClicked()));
        connect(std::get<1>(header), SIGNAL(clicked(bool)), sensorDevices, SLOT(setShown(bool)));
        connect(std::get<1>(header), SIGNAL(clicked(bool)), this, SLOT(changeVisibleCount()));
        widget.verticalLayout->addWidget(std::get<0>(header));
        widget.verticalLayout->addWidget(sensorDevices);
        sensorDevices->setVisible(false);
        sensorDevices->setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::MinimumExpanding);
    }
    //add subwidget LVL1Controllers
    {
        lvl1Controllers = new RobotUnitLVL1ControllersWidget;
        //header
        auto header = makeHeaderItem("LVL1Controllers");
        connect(std::get<2>(header), SIGNAL(clicked(bool)), this, SLOT(refreshLVL1ControllersClicked()));
        connect(std::get<1>(header), SIGNAL(clicked(bool)), lvl1Controllers, SLOT(setShown(bool)));
        connect(std::get<1>(header), SIGNAL(clicked(bool)), this, SLOT(changeVisibleCount()));
        widget.verticalLayout->addWidget(std::get<0>(header));
        widget.verticalLayout->addWidget(lvl1Controllers);
        lvl1Controllers->setVisible(false);
        lvl1Controllers->setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::MinimumExpanding);
    }
    //add subwidget LVL1ControllerClasses
    {
        lvl1ControllerClasses = new RobotUnitLVL1ControllerClassesWidget;
        //header
        auto header = makeHeaderItem("LVL1ControllerClasses");
        connect(std::get<2>(header), SIGNAL(clicked(bool)), this, SLOT(refreshLVL1ControllerClassesClicked()));
        connect(std::get<1>(header), SIGNAL(clicked(bool)), lvl1ControllerClasses, SLOT(setShown(bool)));
        connect(std::get<1>(header), SIGNAL(clicked(bool)), this, SLOT(changeVisibleCount()));
        widget.verticalLayout->addWidget(std::get<0>(header));
        widget.verticalLayout->addWidget(lvl1ControllerClasses);
        lvl1ControllerClasses->setVisible(false);
        lvl1ControllerClasses->setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::MinimumExpanding);
    }
    spacer = new QSpacerItem {0, 0, QSizePolicy::Minimum, QSizePolicy::Expanding};
    changeVisibleCount();
    connect(widget.comboBoxPackage, SIGNAL(currentIndexChanged(QString)), this, SLOT(packageEditChanged()));
    connect(widget.comboBoxPackage, SIGNAL(editTextChanged(QString)), this, SLOT(packageEditChanged()));
    widget.comboBoxPackage->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
    widget.comboBoxPackage->setFixedWidth(300);

    //get package hints
    {
        using namespace boost::filesystem;
        std::string homeDir = QDir::homePath().toStdString();
        path p = path {homeDir} / ".cmake" / "packages";
        if (is_directory(p))
        {
            for (const path & entry : boost::make_iterator_range(directory_iterator(p), {}))
            {
                const std::string pkg = entry.filename().string();
                if (CMakePackageFinder {pkg, "", true} .packageFound())
                {
                    widget.comboBoxPackage->addItem(QString::fromStdString(pkg));
                }
            }
        }
    }

    widget.comboBoxPackage->setEditText("");
    packageEditChanged();

}

RobotUnitPluginWidgetController::~RobotUnitPluginWidgetController()
{
    if (!spacerAdded)
    {
        delete spacer;
    }
}

void RobotUnitPluginWidgetController::loadSettings(QSettings* settings)
{
    robotUnitProxyName = settings->value("robotUnitProxyName", QString::fromStdString(robotUnitProxyName)).toString().toStdString();
}

void RobotUnitPluginWidgetController::saveSettings(QSettings* settings)
{
    settings->setValue("robotUnitProxyName", QString::fromStdString(robotUnitProxyName));
}

void RobotUnitPluginWidgetController::onInitComponent()
{
    usingProxy(robotUnitProxyName);
    ARMARX_INFO << "RobotUnitPluginWidgetController::onInitComponent()" << std::flush;
}

void RobotUnitPluginWidgetController::onConnectComponent()
{
    robotUnitPrx = getProxy<RobotUnitInterfacePrx>(robotUnitProxyName);
    listenerTopicName = robotUnitPrx->getListenerTopicName();
    QMetaObject::invokeMethod(this, "refreshLVL1ControllersClicked", Qt::QueuedConnection);
    QMetaObject::invokeMethod(this, "refreshLVL1ControllerClassesClicked", Qt::QueuedConnection);
    QMetaObject::invokeMethod(this, "refreshControlDevicesClicked", Qt::QueuedConnection);
    QMetaObject::invokeMethod(this, "refreshSensorDevicesClicked", Qt::QueuedConnection);
    usingTopic(listenerTopicName);
}

void RobotUnitPluginWidgetController::onDisconnectComponent()
{
    unsubscribeFromTopic(listenerTopicName);
    robotUnitPrx = nullptr;
    QMetaObject::invokeMethod(this, "refreshLVL1ControllersClicked", Qt::QueuedConnection);
    QMetaObject::invokeMethod(this, "refreshLVL1ControllerClassesClicked", Qt::QueuedConnection);
    QMetaObject::invokeMethod(this, "refreshControlDevicesClicked", Qt::QueuedConnection);
    QMetaObject::invokeMethod(this, "refreshSensorDevicesClicked", Qt::QueuedConnection);
}

QPointer<QDialog> RobotUnitPluginWidgetController::getConfigDialog(QWidget* parent)
{
    if (!dialog)
    {
        dialog = new SimpleConfigDialog(parent);
        dialog->addProxyFinder<RobotUnitInterfacePrx>({"RobotUnit", "The RobotUnit", "*"});
    }
    return qobject_cast<SimpleConfigDialog*>(dialog);
}

void RobotUnitPluginWidgetController::configured()
{
    robotUnitProxyName = dialog->getProxyName("RobotUnit");
}

void RobotUnitPluginWidgetController::lVl1ControllerStatusChanged(const LVL1ControllerStatus& status, const Ice::Current&)
{
    lvl1Controllers->lVl1ControllerStatusChanged(status);
}

void RobotUnitPluginWidgetController::controlDeviceStatusChanged(const ControlDeviceStatus& status, const Ice::Current&)
{
    controlDevices->controlDeviceStatusChanged(status);
}

void RobotUnitPluginWidgetController::sensorDeviceStatusChanged(const SensorDeviceStatus& status, const Ice::Current&)
{
    sensorDevices->sensorDeviceStatusChanged(status);
}

void RobotUnitPluginWidgetController::lvl1ControllerClassAdded(const std::string& name, const Ice::Current&)
{
    QMetaObject::invokeMethod(lvl1ControllerClasses, "lvl1ControllerClassAdded", Qt::QueuedConnection, Q_ARG(std::string, name));
}

void RobotUnitPluginWidgetController::lvl1ControllerCreated(const std::string& name, const Ice::Current&)
{
    QMetaObject::invokeMethod(lvl1Controllers, "lvl1ControllerCreated", Qt::QueuedConnection , Q_ARG(std::string, name));
}

void RobotUnitPluginWidgetController::lvl1ControllerDeleted(const std::string& name, const Ice::Current&)
{
    QMetaObject::invokeMethod(lvl1Controllers, "lvl1ControllerDeleted", Qt::QueuedConnection , Q_ARG(std::string, name));
}

void RobotUnitPluginWidgetController::refreshLVL1ControllersClicked()
{
    lvl1Controllers->reset(robotUnitPrx);
}

void RobotUnitPluginWidgetController::refreshLVL1ControllerClassesClicked()
{
    lvl1ControllerClasses->reset(robotUnitPrx);
}

void RobotUnitPluginWidgetController::refreshControlDevicesClicked()
{
    controlDevices->reset(robotUnitPrx);
}

void RobotUnitPluginWidgetController::refreshSensorDevicesClicked()
{
    sensorDevices->reset(robotUnitPrx);
}

void RobotUnitPluginWidgetController::changeVisibleCount()
{
    if (sensorDevices->isVisible() || lvl1Controllers->isVisible() || lvl1ControllerClasses->isVisible() || controlDevices->isVisible())
    {
        if (spacerAdded)
        {
            spacerAdded = false;
            widget.verticalLayout->removeItem(spacer);
        }
    }
    else
    {
        if (!spacerAdded)
        {
            spacerAdded = true;
            widget.verticalLayout->addItem(spacer);
        }
    }
}

void RobotUnitPluginWidgetController::loadLibClicked()
{
    if (!robotUnitPrx)
    {
        return;
    }

    switch (selectLibMode)
    {
        case SelectLibsMode::LineEdit:
        {
            auto lib = widget.lineEditLibrary->text().toStdString();
            if (lib.empty())
            {
                return;
            }
            ARMARX_INFO << "requesting to load lib " << lib;
            robotUnitPrx->loadLibFromPath(lib);
        }
        break;

        case SelectLibsMode::ComboBox:
        {
            auto package = widget.comboBoxPackage->currentText().toStdString();
            auto lib = widget.comboBoxLibrary->currentText().toStdString();
            if (lib.empty())
            {
                return;
            }
            if (libShortNameToFileName.count(lib))
            {
                lib = libShortNameToFileName.at(lib);
            }
            ARMARX_INFO << "requesting to load lib " << lib << " from package " << package;
            robotUnitPrx->loadLibFromPackage(package, lib);
        }
        break;
        default:
            ARMARX_WARNING << "Load lib  for given SelectLibsMode nyi";
            break;
    }
}

void RobotUnitPluginWidgetController::packageEditChanged()
{
    auto package = widget.comboBoxPackage->currentText();
    if (package.isEmpty())
    {
        selectLibMode = SelectLibsMode::LineEdit;
        widget.lineEditLibrary->setVisible(true);
        widget.comboBoxLibrary->setVisible(false);
        widget.pushButtonLoadLib->setEnabled(true);
        widget.labelPackageFound->setPixmap(QPixmap(":/icons/Blank.svg").scaled(16, 16));
        widget.labelPackageFound->setToolTip("");
    }
    else
    {
        libShortNameToFileName.clear();
        widget.comboBoxLibrary->clear();

        selectLibMode = SelectLibsMode::ComboBox;
        widget.lineEditLibrary->setVisible(false);
        widget.comboBoxLibrary->setVisible(true);
        CMakePackageFinder pFinder(package.toStdString());
        if (pFinder.packageFound())
        {
            widget.pushButtonLoadLib->setEnabled(true);
            widget.labelPackageFound->setPixmap(QPixmap(":/icons/user-online.svg").scaled(16, 16));
            widget.labelPackageFound->setToolTip("Found Package");
            int libidx = -1;
            for (const std::string& lib : Split(pFinder.getLibs(), ",; ", true))
            {
                if (lib.empty())
                {
                    return;
                }
                const auto libPrefix = lib.find("lib");
                const auto libSubstrStart = libPrefix == lib.npos ? 0 : libPrefix + 3;
                const auto libSuffix = lib.find(".");
                const auto libSubstrEnd = libSuffix - libSubstrStart;
                std::string shortName = lib.substr(libSubstrStart, libSubstrEnd);
                libShortNameToFileName[shortName] = lib;
                widget.comboBoxLibrary->addItem(QString::fromStdString(shortName));
                if (libidx == -1 && (lib.find("Controller") != lib.npos || lib.find("controller") != lib.npos))
                {
                    libidx = libShortNameToFileName.size() - 1 ;
                }
                widget.comboBoxLibrary->setCurrentIndex(libidx);
            }
        }
        else
        {
            widget.pushButtonLoadLib->setEnabled(false);
            widget.labelPackageFound->setPixmap(QPixmap(":/icons/dialog-cancel-5.svg").scaled(16, 16));
            widget.labelPackageFound->setToolTip("Cannot find Package");
        }
    }
}

std::tuple<QWidget*, QCheckBox*, QPushButton*> RobotUnitPluginWidgetController::makeHeaderItem(const QString& name)
{
    QWidget* w = new QWidget;
    w->setLayout(new QHBoxLayout);
    w->layout()->setContentsMargins(0, 0, 0, 0);
    w->setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::Minimum);
    QCheckBox* c = new QCheckBox {name};
    c->setStyleSheet(checkboxStyleSheet());
    c->setChecked(false);
    w->layout()->addWidget(c);
    w->layout()->addItem(new QSpacerItem {0, 0, QSizePolicy::MinimumExpanding});
    QPushButton* b = new QPushButton;
    b->setIcon(QIcon {QString{":/icons/view-refresh-7.png"}});
    b->setToolTip("Refreash the List of " + name);
    w->layout()->addWidget(b);
    return std::tuple<QWidget*, QCheckBox*, QPushButton*> {w, c, b};
}
