/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * \package    RobotAPI::gui-plugins::RobotUnitPlugin
 * \author     Raphael ( ufdrv at student dot kit dot edu )
 * \date       2017
 * \copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */
#include "RobotUnitSensorDevicesWidget.h"
#include "ui_RobotUnitSensorDevicesWidget.h"

#include "StyleSheets.h"

#include <QCheckBox>

namespace armarx
{
    RobotUnitSensorDevicesWidget::RobotUnitSensorDevicesWidget(QWidget* parent) :
        QWidget(parent),
        ui(new Ui::RobotUnitSensorDevicesWidget)
    {
        ui->setupUi(this);
        ui->treeWidget->setColumnCount(4);
        QTreeWidgetItem* head = ui->treeWidget->headerItem();
        head->setText(idxName, "Name");
        head->setText(idxTags, "Tags");
        head->setText(idxValType, "Value Taype");
        head->setText(idxValue, "Value");
        head->setToolTip(idxName, "The control device's name");
        head->setToolTip(idxTags, "The control device's tags");
        head->setToolTip(idxValType, "The sensor value's' type");
        head->setToolTip(idxValue, "The sensor value");
        ui->treeWidget->setColumnWidth(idxName, 150);
        ui->treeWidget->setColumnWidth(idxTags, 100);
        ui->treeWidget->setColumnWidth(idxValType, 450);
    }

    RobotUnitSensorDevicesWidget::~RobotUnitSensorDevicesWidget()
    {
        delete ui;
    }

    void RobotUnitSensorDevicesWidget::sensorDeviceStatusChanged(const SensorDeviceStatus& status)
    {
        std::lock_guard<std::recursive_mutex> guard {statusUpdatesMutex};
        if (statusUpdates[status.deviceName].timestampUSec < status.timestampUSec)
        {
            statusUpdates[status.deviceName] = status;
            QMetaObject::invokeMethod(this, "sensorDeviceStatusChanged", Qt::QueuedConnection);
        }
    }

    void RobotUnitSensorDevicesWidget::sensorDeviceStatusChanged()
    {
        std::lock_guard<std::recursive_mutex> guard {statusUpdatesMutex};
        if (!robotUnit)
        {
            return;
        }
        for (const auto& pair : statusUpdates)
        {
            if (!entries.count(pair.second.deviceName))
            {
                add(robotUnit->getSensorDeviceDescription(pair.second.deviceName));
            }
            entries.at(pair.second.deviceName)->update(pair.second);
        }
        statusUpdates.clear();
    }

    void RobotUnitSensorDevicesWidget::reset(RobotUnitInterfacePrx ru)
    {
        std::lock_guard<std::recursive_mutex> guard {mutex};
        robotUnit = ru;
        //delete all
        ui->treeWidget->clear();
        entries.clear();
        //fill
        if (robotUnit)
        {
            for (SensorDeviceDescription& desc : robotUnit->getSensorDeviceDescriptions())
            {
                add(desc);
            }
        }
    }

    void RobotUnitSensorDevicesWidget::add(const SensorDeviceDescription& desc)
    {
        std::lock_guard<std::recursive_mutex> guard {mutex};
        if (!entries.count(desc.deviceName))
        {
            entries[desc.deviceName] = new RobotUnitSensorDevicesWidgetEntry(*this, *(ui->treeWidget), desc);
        }
    }

    RobotUnitSensorDevicesWidgetEntry::RobotUnitSensorDevicesWidgetEntry(RobotUnitSensorDevicesWidget& parent, QTreeWidget& treeWidget, const SensorDeviceDescription& desc):
        QObject {&parent}
    {
        header = new QTreeWidgetItem;
        treeWidget.addTopLevelItem(header);
        header->setText(RobotUnitSensorDevicesWidget::idxName, QString::fromStdString(desc.deviceName));
        //tags
        {
            QCheckBox* boxTags = new QCheckBox{QString::number(desc.tags.size()) + " Tags"};
            treeWidget.setItemWidget(header, RobotUnitSensorDevicesWidget::idxTags, boxTags);
            boxTags->setStyleSheet(checkboxStyleSheet());
            boxTags->setChecked(false);
            connect(boxTags, SIGNAL(clicked(bool)), this, SLOT(hideTagList(bool)));
            for (const auto& tag : desc.tags)
            {
                QTreeWidgetItem* child = new QTreeWidgetItem {{QString::fromStdString(tag)}};
                treeWidget.addTopLevelItem(child);
                tags.emplace_back(child);
            }
        }
        header->setText(RobotUnitSensorDevicesWidget::idxValType, QString::fromStdString(desc.sensorValueType));
        value = new VariantWidget;
        QTreeWidgetItem* valchild = new QTreeWidgetItem;
        header->addChild(valchild);
        treeWidget.setItemWidget(valchild, RobotUnitSensorDevicesWidget::idxValue, value);
    }

    void RobotUnitSensorDevicesWidgetEntry::update(const SensorDeviceStatus& status)
    {
        value->update(status.sensorValue);
        value->layout()->setContentsMargins(0, 0, 0, 0);
    }

    bool RobotUnitSensorDevicesWidgetEntry::matchName(const QString& name)
    {
        return header->text(RobotUnitSensorDevicesWidget::idxName).contains(name, Qt::CaseInsensitive);
    }

    bool RobotUnitSensorDevicesWidgetEntry::matchTag(const QString& tag)
    {
        for (const auto& tagit : tags)
        {
            if (tagit->text(RobotUnitSensorDevicesWidget::idxTags).contains(tag, Qt::CaseInsensitive))
            {
                return true;
            }
        }
        return false;
    }

    bool RobotUnitSensorDevicesWidgetEntry::matchValueType(const QString& type)
    {
        return header->text(RobotUnitSensorDevicesWidget::idxValType).contains(type, Qt::CaseInsensitive);
    }

    void RobotUnitSensorDevicesWidgetEntry::setVisible(bool vis)
    {
        if (!vis)
        {
            hideTagList(true);
        }
        header->setHidden(!vis);
    }

    void RobotUnitSensorDevicesWidgetEntry::hideTagList(bool hide)
    {
        for (QTreeWidgetItem* tag : tags)
        {
            tag->setHidden(hide);
        }
    }

}
